"""Tool for generating an S matrix automatically from a Tidy3d simulation and lumped port definitions."""

from __future__ import annotations

from typing import Dict, Tuple, Union

import numpy as np
import pydantic.v1 as pd

from ....components.base import cached_property
from ....components.data.data_array import (
    DataArray,
    FreqDataArray,
)
from ....components.data.monitor_data import (
    MonitorData,
)
from ....components.data.sim_data import SimulationData
from ....components.geometry.utils_2d import snap_coordinate_to_grid
from ....components.microwave.data.monitor_data import AntennaMetricsData
from ....components.monitor import DirectivityMonitor
from ....components.simulation import Simulation
from ....components.source.time import GaussianPulse
from ....components.types import Ax
from ....components.viz import add_ax_if_none, equal_aspect
from ....constants import C_0, OHM
from ....exceptions import Tidy3dError, Tidy3dKeyError, ValidationError
from ....web.api.container import BatchData
from ..data.terminal import PortDataArray, TerminalPortDataArray
from ..ports.base_lumped import AbstractLumpedPort
from ..ports.coaxial_lumped import CoaxialLumpedPort
from ..ports.rectangular_lumped import LumpedPort
from ..ports.wave import WavePort
from .base import FWIDTH_FRAC, AbstractComponentModeler, TerminalPortType


class TerminalComponentModeler(AbstractComponentModeler):
    """Tool for modeling two-terminal multiport devices and computing port parameters
    with lumped and wave ports."""

    ports: Tuple[TerminalPortType, ...] = pd.Field(
        (),
        title="Terminal Ports",
        description="Collection of lumped and wave ports associated with the network. "
        "For each port, one simulation will be run with a source that is associated with the port.",
    )

    radiation_monitors: tuple[DirectivityMonitor, ...] = pd.Field(
        (),
        title="Radiation Monitors",
        description="Facilitates the calculation of figures-of-merit for antennas. "
        "These monitor will be included in every simulation and record the radiated fields. ",
    )

    @equal_aspect
    @add_ax_if_none
    def plot_sim(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **kwargs
    ) -> Ax:
        """Plot a :class:`.Simulation` with all sources added for each port, for troubleshooting."""

        plot_sources = []
        for port_source in self.ports:
            source_0 = port_source.to_source(self._source_time)
            plot_sources.append(source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot(x=x, y=y, z=z, ax=ax, **kwargs)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **kwargs
    ) -> Ax:
        """Plot permittivity of the :class:`.Simulation` with all sources added for each port."""

        plot_sources = []
        for port_source in self.ports:
            source_0 = port_source.to_source(self._source_time)
            plot_sources.append(source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    @cached_property
    def sim_dict(self) -> Dict[str, Simulation]:
        """Generate all the :class:`.Simulation` objects for the port parameter calculation."""

        sim_dict = {}

        # internal mesh override and snapping points are automatically generated from lumped elements.
        lumped_resistors = [port.to_load() for port in self._lumped_ports]

        # Apply the highest frequency in the simulation to define the grid, rather than the
        # source's central frequency, to ensure an accurate solution over the entire range
        grid_spec = self.simulation.grid_spec.copy(
            update={
                "wavelength": C_0 / np.max(self.freqs),
            }
        )

        # Make an initial simulation with new grid_spec to determine where LumpedPorts are snapped
        sim_wo_source = self.simulation.updated_copy(
            grid_spec=grid_spec, lumped_elements=lumped_resistors
        )
        snap_centers = dict()
        for port in self._lumped_ports:
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                sim_wo_source.grid, port_center_on_axis, port.injection_axis
            )
            snap_centers[port.name] = new_port_center

        # Create monitors and snap to the center positions
        field_monitors = [
            mon
            for port in self.ports
            for mon in port.to_monitors(
                self.freqs, snap_center=snap_centers.get(port.name), grid=sim_wo_source.grid
            )
        ]

        new_mnts = list(self.simulation.monitors) + field_monitors

        if self.radiation_monitors is not None:
            new_mnts = new_mnts + list(self.radiation_monitors)

        new_lumped_elements = list(self.simulation.lumped_elements) + [
            port.to_load(snap_center=snap_centers[port.name]) for port in self._lumped_ports
        ]

        update_dict = dict(
            monitors=new_mnts,
            lumped_elements=new_lumped_elements,
        )

        # This is the new default simulation will all shared components added
        sim_wo_source = sim_wo_source.copy(update=update_dict)

        # Next, simulations are generated that include the source corresponding with the excitation port
        for port in self._lumped_ports:
            port_source = port.to_source(
                self._source_time, snap_center=snap_centers[port.name], grid=sim_wo_source.grid
            )
            task_name = self._task_name(port=port)
            sim_dict[task_name] = sim_wo_source.updated_copy(sources=[port_source])

        # Check final simulations for grid size at lumped ports
        for _, sim in sim_dict.items():
            TerminalComponentModeler._check_grid_size_at_ports(sim, self._lumped_ports)

        # Now, create simulations with wave port sources and mode solver monitors for computing port modes
        for wave_port in self._wave_ports:
            # Source is placed just before the field monitor of the port
            mode_src_pos = wave_port.center[wave_port.injection_axis] + self._shift_value_signed(
                wave_port
            )
            port_source = wave_port.to_source(self._source_time, snap_center=mode_src_pos)

            update_dict = dict(sources=[port_source])

            task_name = self._task_name(port=wave_port)
            sim_dict[task_name] = sim_wo_source.copy(update=update_dict)
        return sim_dict

    @cached_property
    def _source_time(self):
        """Helper to create a time domain pulse for the frequency range of interest."""
        freq0 = np.mean(self.freqs)
        fdiff = max(self.freqs) - min(self.freqs)
        fwidth = max(fdiff, freq0 * FWIDTH_FRAC)
        return GaussianPulse(
            freq0=freq0, fwidth=fwidth, remove_dc_component=self.remove_dc_component
        )

    def _construct_smatrix(self) -> TerminalPortDataArray:
        """Post process :class:`.BatchData` to generate scattering matrix."""
        return self._internal_construct_smatrix(batch_data=self.batch_data)

    def _internal_construct_smatrix(self, batch_data: BatchData) -> TerminalPortDataArray:
        """Post process :class:`.BatchData` to generate scattering matrix, for internal use only."""

        port_names = [port.name for port in self.ports]

        values = np.zeros(
            (len(self.freqs), len(port_names), len(port_names)),
            dtype=complex,
        )
        coords = dict(
            f=np.array(self.freqs),
            port_out=port_names,
            port_in=port_names,
        )
        a_matrix = TerminalPortDataArray(values, coords=coords)
        b_matrix = a_matrix.copy(deep=True)

        # Tabulate the reference impedances at each port and frequency
        port_impedances = self._port_reference_impedances(batch_data=batch_data)

        # loop through source ports
        for port_in in self.ports:
            sim_data = batch_data[self._task_name(port=port_in)]
            a, b = self.compute_power_wave_amplitudes_at_each_port(port_impedances, sim_data)
            indexer = dict(f=a.f, port_in=port_in.name, port_out=a.port)
            a_matrix.loc[indexer] = a
            b_matrix.loc[indexer] = b

        s_matrix = self.ab_to_s(a_matrix, b_matrix)
        return s_matrix

    @pd.validator("simulation")
    def _validate_3d_simulation(cls, val):
        """Error if :class:`.Simulation` is not a 3D simulation"""

        if val.size.count(0.0) > 0:
            raise ValidationError(
                f"'{cls.__name__}' must be setup with a 3D simulation with all sizes greater than 0."
            )
        return val

    @pd.validator("radiation_monitors")
    def _validate_radiation_monitors(cls, val, values):
        freqs = set(values.get("freqs"))
        for rad_mon in val:
            mon_freqs = rad_mon.freqs
            is_subset = freqs.issuperset(mon_freqs)
            if not is_subset:
                raise ValidationError(
                    f"The frequencies in the radiation monitor '{rad_mon.name}' "
                    f"must be equal to or a subset of the frequencies in the '{cls.__name__}'."
                )
        return val

    @staticmethod
    def _check_grid_size_at_ports(simulation: Simulation, ports: list[Union[AbstractLumpedPort]]):
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        yee_grid = simulation.grid.yee
        for port in ports:
            port._check_grid_size(yee_grid)

    def compute_power_wave_amplitudes_at_each_port(
        self, port_reference_impedances: PortDataArray, sim_data: SimulationData
    ) -> tuple[PortDataArray, PortDataArray]:
        """Compute the incident and reflected power wave amplitudes at each port.
        The computed amplitudes have not been normalized.

        Parameters
        ----------
        port_reference_impedances : :class:`.PortDataArray`
            Reference impedance at each port.
        sim_data : :class:`.SimulationData`
            Results from the simulation.

        Returns
        -------
        tuple[:class:`.PortDataArray`, :class:`.PortDataArray`]
            Incident (a) and reflected (b) power wave amplitudes at each port.
        """
        port_names = [port.name for port in self.ports]
        values = np.zeros(
            (len(self.freqs), len(port_names)),
            dtype=complex,
        )
        coords = dict(
            f=np.array(self.freqs),
            port=port_names,
        )

        V_matrix = PortDataArray(values, coords=coords)
        I_matrix = V_matrix.copy(deep=True)
        a = V_matrix.copy(deep=True)
        b = V_matrix.copy(deep=True)

        for port_out in self.ports:
            V_out, I_out = self.compute_port_VI(port_out, sim_data)
            indexer = dict(port=port_out.name)
            V_matrix.loc[indexer] = V_out
            I_matrix.loc[indexer] = I_out

        V_numpy = V_matrix.values
        I_numpy = I_matrix.values
        Z_numpy = port_reference_impedances.values

        # Check to make sure sign is consistent for all impedance values
        self._check_port_impedance_sign(Z_numpy)

        # # Check for negative real part of port impedance and flip the V and Z signs accordingly
        negative_real_Z = np.real(Z_numpy) < 0
        V_numpy = np.where(negative_real_Z, -V_numpy, V_numpy)
        Z_numpy = np.where(negative_real_Z, -Z_numpy, Z_numpy)

        F_numpy = TerminalComponentModeler._compute_F(Z_numpy)

        # Equation 4.67 - Pozar - Microwave Engineering 4ed
        a.values = F_numpy * (V_numpy + Z_numpy * I_numpy)
        b.values = F_numpy * (V_numpy - np.conj(Z_numpy) * I_numpy)

        return a, b

    @staticmethod
    def compute_port_VI(
        port_out: TerminalPortType, sim_data: SimulationData
    ) -> tuple[FreqDataArray, FreqDataArray]:
        """Compute the port voltages and currents.

        Parameters
        ----------
        port_out : ``TerminalPortType``
            Port for computing voltage and current.
        sim_data : :class:`.SimulationData`
            Results from simulation containing field data.

        Returns
        -------
        tuple[FreqDataArray, FreqDataArray]
            Voltage and current values at the port as frequency arrays.
        """
        voltage = port_out.compute_voltage(sim_data)
        current = port_out.compute_current(sim_data)
        return voltage, current

    @staticmethod
    def compute_power_wave_amplitudes(
        port: Union[LumpedPort, CoaxialLumpedPort], sim_data: SimulationData
    ) -> tuple[FreqDataArray, FreqDataArray]:
        """Compute the incident and reflected power wave amplitudes at a lumped port.
        The computed amplitudes have not been normalized.

        Parameters
        ----------
        port : Union[:class:`.LumpedPort`, :class:`.CoaxialLumpedPort`]
            Port for computing voltage and current.
        sim_data : :class:`.SimulationData`
            Results from the simulation.

        Returns
        -------
        tuple[FreqDataArray, FreqDataArray]
            Incident (a) and reflected (b) power wave amplitude frequency arrays.
        """
        voltage, current = TerminalComponentModeler.compute_port_VI(port, sim_data)
        # Amplitudes for the incident and reflected power waves
        a = (voltage + port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
        b = (voltage - port.impedance * current) / 2 / np.sqrt(np.real(port.impedance))
        return a, b

    @staticmethod
    def compute_power_delivered_by_port(
        port: Union[LumpedPort, CoaxialLumpedPort], sim_data: SimulationData
    ) -> FreqDataArray:
        """Compute the power delivered to the network by a lumped port.

        Parameters
        ----------
        port : Union[:class:`.LumpedPort`, :class:`.CoaxialLumpedPort`]
            Port for computing voltage and current.
        sim_data : :class:`.SimulationData`
            Results from the simulation.

        Returns
        -------
        FreqDataArray
            Power in units of Watts as a frequency array.
        """
        a, b = TerminalComponentModeler.compute_power_wave_amplitudes(sim_data=sim_data, port=port)
        # Power delivered is the incident power minus the reflected power
        return 0.5 * (np.abs(a) ** 2 - np.abs(b) ** 2)

    @staticmethod
    def ab_to_s(
        a_matrix: TerminalPortDataArray, b_matrix: TerminalPortDataArray
    ) -> TerminalPortDataArray:
        """Get the scattering matrix given the power wave matrices."""
        # Ensure dimensions are ordered properly
        a_matrix = a_matrix.transpose(*TerminalPortDataArray._dims)
        b_matrix = b_matrix.transpose(*TerminalPortDataArray._dims)

        s_matrix = a_matrix.copy(deep=True)
        a_vals = s_matrix.copy(deep=True).values
        b_vals = b_matrix.copy(deep=True).values

        s_vals = np.matmul(b_vals, AbstractComponentModeler.inv(a_vals))

        s_matrix.data = s_vals
        return s_matrix

    @staticmethod
    def s_to_z(
        s_matrix: TerminalPortDataArray, reference: Union[complex, PortDataArray]
    ) -> DataArray:
        """Get the impedance matrix given the scattering matrix and a reference impedance."""

        # Ensure dimensions are ordered properly
        z_matrix = s_matrix.transpose(*TerminalPortDataArray._dims).copy(deep=True)
        s_vals = z_matrix.values
        eye = np.eye(len(s_matrix.port_out.values), len(s_matrix.port_in.values))
        if isinstance(reference, PortDataArray):
            # From Equation 4.68 - Pozar - Microwave Engineering 4ed
            # Ensure that Zport, F, and Finv act as diagonal matrices when multiplying by left or right
            shape_left = (len(s_matrix.f), len(s_matrix.port_out), 1)
            shape_right = (len(s_matrix.f), 1, len(s_matrix.port_in))
            Zport = reference.values.reshape(shape_right)
            F = TerminalComponentModeler._compute_F(Zport).reshape(shape_right)
            Finv = (1.0 / F).reshape(shape_left)
            FinvSF = Finv * s_vals * F
            RHS = eye * np.conj(Zport) + FinvSF * Zport
            LHS = eye - FinvSF
            z_vals = np.matmul(AbstractComponentModeler.inv(LHS), RHS)
        else:
            # Simpler case when all port impedances are the same
            z_vals = (
                np.matmul(AbstractComponentModeler.inv(eye - s_vals), (eye + s_vals)) * reference
            )

        z_matrix.data = z_vals
        return z_matrix

    @cached_property
    def port_reference_impedances(self) -> PortDataArray:
        """The reference impedance used at each port for definining power wave amplitudes."""
        return self._port_reference_impedances(self.batch_data)

    def _port_reference_impedances(self, batch_data: BatchData) -> PortDataArray:
        """Tabulates the reference impedance of each port at each frequency using the
        supplied :class:`.BatchData`.
        """
        port_names = [port.name for port in self.ports]

        values = np.zeros(
            (len(self.freqs), len(port_names)),
            dtype=complex,
        )
        coords = dict(f=np.array(self.freqs), port=port_names)
        port_impedances = PortDataArray(values, coords=coords)
        for port in self.ports:
            if isinstance(port, WavePort):
                # Mode solver data for each wave port is stored in its associated SimulationData
                sim_data_port = batch_data[self._task_name(port=port)]
                # WavePorts have a port impedance calculated from its associated modal field distribution
                # and is frequency dependent.
                impedances = port.compute_port_impedance(sim_data_port).values
                port_impedances.loc[dict(port=port.name)] = impedances.squeeze()
            else:
                # LumpedPorts have a constant reference impedance
                port_impedances.loc[dict(port=port.name)] = np.full(len(self.freqs), port.impedance)

        port_impedances = TerminalComponentModeler._set_port_data_array_attributes(port_impedances)
        return port_impedances

    @staticmethod
    def _compute_F(Z_numpy: np.array):
        """Helper to convert port impedance matrix to F, which is used for
        computing generalized scattering parameters."""
        return 1.0 / (2.0 * np.sqrt(np.real(Z_numpy)))

    @cached_property
    def _lumped_ports(self) -> list[AbstractLumpedPort]:
        """A list of all lumped ports in the ``TerminalComponentModeler``"""
        return [port for port in self.ports if isinstance(port, AbstractLumpedPort)]

    @cached_property
    def _wave_ports(self) -> list[WavePort]:
        """A list of all wave ports in the ``TerminalComponentModeler``"""
        return [port for port in self.ports if isinstance(port, WavePort)]

    @staticmethod
    def _set_port_data_array_attributes(data_array: PortDataArray) -> PortDataArray:
        """Helper to set additional metadata for ``PortDataArray``."""
        data_array.name = "Z0"
        return data_array.assign_attrs(units=OHM, long_name="characteristic impedance")

    @staticmethod
    def _check_port_impedance_sign(Z_numpy: np.ndarray):
        """Sanity check for consistent sign of real part of Z for each port across all frequencies."""
        for port_idx in range(Z_numpy.shape[1]):
            port_Z = Z_numpy[:, port_idx]
            signs = np.sign(np.real(port_Z))
            if not np.all(signs == signs[0]):
                raise Tidy3dError(
                    f"Inconsistent sign of real part of Z detected for port {port_idx}. "
                    "If you received this error, please create an issue in the Tidy3D "
                    "github repository."
                )

    def get_radiation_monitor_by_name(self, monitor_name: str) -> DirectivityMonitor:
        """Find and return a :class:`.DirectivityMonitor` monitor by its name.

        Parameters
        ----------
        monitor_name : str
            Name of the monitor to find.

        Returns
        -------
        :class:`.DirectivityMonitor`
            The monitor matching the given name.

        Raises
        ------
        ``Tidy3dKeyError``
            If no monitor with the given name exists.
        """
        for monitor in self.radiation_monitors:
            if monitor.name == monitor_name:
                return monitor
        raise Tidy3dKeyError(f"No radiation monitor named '{monitor_name}'.")

    def _monitor_data_at_port_amplitude(
        self,
        port: TerminalPortType,
        sim_data: SimulationData,
        monitor_data: MonitorData,
        a_port: Union[FreqDataArray, complex],
    ) -> MonitorData:
        """Normalize the monitor data to a desired complex amplitude of a port,
        represented by ``a_port``, where :math:`\\frac{1}{2}|a|^2` is the power
        incident from the port into the system.
        """
        a_raw, _ = self.compute_power_wave_amplitudes_at_each_port(
            self.port_reference_impedances, sim_data
        )
        a_raw_port = a_raw.sel(port=port.name)
        if not isinstance(a_port, FreqDataArray):
            freqs = list(monitor_data.monitor.freqs)
            array_vals = a_port * np.ones(len(freqs))
            a_port = FreqDataArray(array_vals, coords=dict(f=freqs))
        scale_array = a_port / a_raw_port
        return monitor_data.scale_fields_by_freq_array(scale_array, method="nearest")

    def get_antenna_metrics_data(
        self, port_amplitudes: dict[str, complex] = None, monitor_name: str = None
    ) -> AntennaMetricsData:
        """Calculate antenna parameters using superposition of fields from multiple port excitations.

        The method computes the radiated far fields and port excitation power wave amplitudes
        for a superposition of port excitations, which can be used to analyze antenna radiation
        characteristics.

        Parameters
        ----------
        port_amplitudes : dict[str, complex] = None
            Dictionary mapping port names to their desired excitation amplitudes. For each port,
            :math:`\\frac{1}{2}|a|^2` represents the incident power from that port into the system.
            If None, uses only the first port without any scaling of the raw simulation data.
        monitor_name : str = None
            Name of the :class:`.DirectivityMonitor` to use for calculating far fields.
            If None, uses the first monitor in `radiation_monitors`.

        Returns
        -------
        :class:`.AntennaMetricsData`
            Container with antenna parameters including directivity, gain, and radiation efficiency,
            computed from the superposition of fields from all excited ports.
        """
        # Use the first port as default if none specified
        if port_amplitudes is None:
            port_amplitudes = {self.ports[0].name: None}
        port_names = [port.name for port in self.ports]
        # Check port names, and create map from port to amplitude
        port_dict = {}
        for key in port_amplitudes.keys():
            port = self.get_port_by_name(port_name=key)
            port_dict[port] = port_amplitudes[key]
        # Get the radiation monitor, use first as default
        # if none specified
        if monitor_name is None:
            rad_mon = self.radiation_monitors[0]
        else:
            rad_mon = self.get_radiation_monitor_by_name(monitor_name)

        # Create data arrays for holding the superposition of all port power wave amplitudes
        f = list(rad_mon.freqs)
        coords = dict(f=f, port=port_names)
        a_sum = PortDataArray(np.zeros((len(f), len(port_names)), dtype=complex), coords=coords)
        b_sum = a_sum.copy()
        # Retrieve associated simulation data
        combined_directivity_data = None
        for port, amplitude in port_dict.items():
            sim_data_port = self.batch_data[self._task_name(port=port)]
            radiation_data = sim_data_port[rad_mon.name]

            a, b = self.compute_power_wave_amplitudes_at_each_port(
                self.port_reference_impedances, sim_data_port
            )
            # Select a possible subset of frequencies
            a = a.sel(f=f)
            b = b.sel(f=f)
            a_raw = a.sel(port=port.name)

            if amplitude is None:
                # No scaling performed when amplitude is None
                scaled_directivity_data = sim_data_port[rad_mon.name]
                scale_factor = 1.0
            else:
                scaled_directivity_data = self._monitor_data_at_port_amplitude(
                    port, sim_data_port, radiation_data, amplitude
                )
                scale_factor = amplitude / a_raw
            a = scale_factor * a
            b = scale_factor * b

            # Combine the possibly scaled directivity data and the power wave amplitudes
            if combined_directivity_data is None:
                combined_directivity_data = scaled_directivity_data
            else:
                combined_directivity_data = combined_directivity_data + scaled_directivity_data
            a_sum += a
            b_sum += b

        # Compute and add power measures to results
        power_incident = np.real(0.5 * a_sum * np.conj(a_sum)).sum(dim="port")
        power_reflected = np.real(0.5 * b_sum * np.conj(b_sum)).sum(dim="port")
        return AntennaMetricsData.from_directivity_data(
            combined_directivity_data, power_incident, power_reflected
        )



# utilities for autograd derivative passing
from __future__ import annotations

import numpy as np
import pydantic.v1 as pd
import xarray as xr

from ...constants import LARGE_NUMBER
from ..base import Tidy3dBaseModel
from ..data.data_array import ScalarFieldDataArray, SpatialDataArray
from ..types import ArrayLike, Bound, tidycomplex
from .types import PathType
from .utils import get_static

# we do this because importing these creates circular imports
FieldData = dict[str, ScalarFieldDataArray]
PermittivityData = dict[str, ScalarFieldDataArray]


class DerivativeSurfaceMesh(Tidy3dBaseModel):
    """Stores information about the surfaces of an object to be used for derivative calculation.

    User guide: this class is used to construct derivatives with respect to surface elements
    given some forward and adjoint fields stored in a ``DerivativeInfo`` class. To use it,
    you must specify the central locations of all the surface elements in ``centers``,
    along with the area of each element in ``areas``. Then you need to specify three orthogonal
    vectors (``normals``, ``perps1`` and ``perps2``). The derivative will be computed with
    respect to a change in material along the ``normals`` direction. It is important to note
    that the sign of these basis vectors is irrelevant since we end up multiplying the forward
    and adjoint fields together in these bases.

    The gradient is with respect to a change from the surface element permittivity from the
    ``eps_in`` to ``eps_out`` field. So in principle it is always computing gradients with
    respect to an infinitesimal outward shift in the normal direction. For example, for
    ``PolySlab.vertices``, this means shifting each edge to the background.
    For ``PolySlab.slab_bounds``, this means shifting the bound in either + or - direction if it
    is index ``[1]`` or ``[0]`` respectively. This renders the sign of the normal vector
    irrelevant.

    After this ``DerivativeSurfaceMesh`` is constructed, given some ``DerivativeInfo`` in the
    gradient calculation, a call to ``DerivativeInfo.grad_surfaces(x: DerivativeSurfaceMesh)``
    will return the gradient with respect to a change in all of the provided surfaces.

    """

    centers: ArrayLike = pd.Field(
        ...,
        title="Centers",
        description="(N, 3) array storing the centers of each surface element.",
    )

    areas: ArrayLike = pd.Field(
        ...,
        title="Area Elements",
        description="(N,) array storing the first perpendicular vectors of each surface element.",
    )

    normals: ArrayLike = pd.Field(
        ...,
        title="Normals",
        description="(N, 3) array storing the normal vectors of each surface element.",
    )

    perps1: ArrayLike = pd.Field(
        ...,
        title="Perpendiculars 1",
        description="(N, 3) array storing the first perpendicular vectors of each surface element.",
    )

    perps2: ArrayLike = pd.Field(
        ...,
        title="Perpendiculars 1",
        description="(N, 3) array storing the first perpendicular vectors of each surface element.",
    )


class DerivativeInfo(Tidy3dBaseModel):
    """Stores derivative information passed to the ``.compute_derivatives`` methods."""

    paths: list[PathType] = pd.Field(
        ...,
        title="Paths to Traced Fields",
        description="List of paths to the traced fields that need derivatives calculated.",
    )

    E_der_map: FieldData = pd.Field(
        ...,
        title="Electric Field Gradient Map",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` store the '
        "multiplication of the forward and adjoint electric fields. The tangential components "
        "of this dataset is used when computing adjoint gradients for shifting boundaries. "
        "All components are used when computing volume-based gradients.",
    )

    D_der_map: FieldData = pd.Field(
        ...,
        title="Displacement Field Gradient Map",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` store the '
        "multiplication of the forward and adjoint displacement fields. The normal component "
        "of this dataset is used when computing adjoint gradients for shifting boundaries.",
    )

    E_fwd: FieldData = pd.Field(
        ...,
        title="Forward Electric Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "forward electric fields used for computing gradients for a given structure.",
    )

    E_adj: FieldData = pd.Field(
        ...,
        title="Adjoint Electric Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "adjoint electric fields used for computing gradients for a given structure.",
    )

    D_fwd: FieldData = pd.Field(
        ...,
        title="Forward Displacement Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "forward displacement fields used for computing gradients for a given structure.",
    )

    D_adj: FieldData = pd.Field(
        ...,
        title="Adjoint Displacement Fields",
        description='Dataset where the field components ``("Ex", "Ey", "Ez")`` represent the '
        "adjoint displacement fields used for computing gradients for a given structure.",
    )

    eps_data: PermittivityData = pd.Field(
        ...,
        title="Permittivity Dataset",
        description="Dataset of relative permittivity values along all three dimensions. "
        "Used for automatically computing permittivity inside or outside of a simple geometry.",
    )

    eps_in: tidycomplex = pd.Field(
        title="Permittivity Inside",
        description="Permittivity inside of the ``Structure``. "
        "Typically computed from ``Structure.medium.eps_model``."
        "Used when it can not be computed from ``eps_data`` or when ``eps_approx==True``.",
    )

    eps_out: tidycomplex = pd.Field(
        ...,
        title="Permittivity Outside",
        description="Permittivity outside of the ``Structure``. "
        "Typically computed from ``Simulation.medium.eps_model``."
        "Used when it can not be computed from ``eps_data`` or when ``eps_approx==True``.",
    )

    eps_background: tidycomplex = pd.Field(
        None,
        title="Permittivity in Background",
        description="Permittivity outside of the ``Structure`` as manually specified by. "
        "``Structure.background_medium``. ",
    )

    bounds: Bound = pd.Field(
        ...,
        title="Geometry Bounds",
        description="Bounds corresponding to the structure, used in ``Medium`` calculations.",
    )

    bounds_intersect: Bound = pd.Field(
        ...,
        title="Geometry and Simulation Intersections Bounds",
        description="Bounds corresponding to the minimum intersection between the "
        "structure and the simulation it is contained in.",
    )

    frequency: float = pd.Field(
        ...,
        title="Frequency of adjoint simulation",
        description="Frequency at which the adjoint gradient is computed.",
    )

    eps_no_structure: SpatialDataArray = pd.Field(
        None,
        title="Permittivity Without Structure",
        description="The permittivity of the original simulation without the structure that is "
        "being differentiated with respect to. Used to approximate permittivity outside of the "
        "structure for shape optimization.",
    )

    eps_inf_structure: SpatialDataArray = pd.Field(
        None,
        title="Permittivity With Infinite Structure",
        description="The permittivity of the original simulation where the structure being "
        " differentiated with respect to is inifinitely large. Used to approximate permittivity "
        "inside of the structure for shape optimization.",
    )

    eps_approx: bool = pd.Field(
        False,
        title="Use Permittivity Approximation",
        description="If ``True``, approximates outside permittivity using ``Simulation.medium``"
        "and the inside permittivity using ``Structure.medium``. "
        "Only set ``True`` for ``GeometryGroup`` handling where it is difficult to automatically "
        "evaluate the inside and outside relative permittivity for each geometry.",
    )

    def updated_paths(self, paths: list[PathType]) -> DerivativeInfo:
        """Update this ``DerivativeInfo`` with new set of paths."""
        return self.updated_copy(paths=paths)

    def _eps_in(self, spatial_coords: np.ndarray) -> complex | np.ndarray:
        """permittivity inside, used internally."""
        # determine inside medium
        if self.eps_inf_structure is not None:
            return self.evaluate_eps(spatial_coords, is_inside=True)

        return self.eps_in

    def _eps_out(self, spatial_coords: np.ndarray) -> complex | np.ndarray:
        """permittivity outside, used internally."""

        # determine background medium
        if self.eps_background is not None:
            return self.eps_background

        if self.eps_no_structure is not None:
            return self.evaluate_eps(spatial_coords, is_inside=False)

        return self.eps_out

    def delta_eps(self, spatial_coords: np.ndarray) -> complex | np.ndarray:
        """Change in the permittivity across interface (for E field grads)."""
        return self._eps_in(spatial_coords) - self._eps_out(spatial_coords)

    def delta_eps_inv(self, spatial_coords: np.ndarray) -> complex | np.ndarray:
        """Change in 1 / permittivity across interface (for D field grads)."""
        return 1.0 / self._eps_in(spatial_coords) - 1.0 / self._eps_out(spatial_coords)

    def grad_surfaces(self, surface_mesh: DerivativeSurfaceMesh) -> dict:
        """Derivative with respect to the surface mesh elements, given the derivative fields."""

        # strip out relevant info from `surface_mesh`
        spatial_coords = surface_mesh.centers
        normals = surface_mesh.normals
        perps1 = surface_mesh.perps1
        perps2 = surface_mesh.perps2

        # unpack electric and displacement fields
        E_fwd = self.E_fwd
        E_adj = self.E_adj
        D_fwd = self.D_fwd
        D_adj = self.D_adj

        # compute the E and D fields at the edge centers
        E_fwd_at_coords = self.evaluate_flds_at(fld_dataset=E_fwd, spatial_coords=spatial_coords)
        E_adj_at_coords = self.evaluate_flds_at(fld_dataset=E_adj, spatial_coords=spatial_coords)
        D_fwd_at_coords = self.evaluate_flds_at(fld_dataset=D_fwd, spatial_coords=spatial_coords)
        D_adj_at_coords = self.evaluate_flds_at(fld_dataset=D_adj, spatial_coords=spatial_coords)

        # project the relevant field quantities into their respective basis for gradient calculation
        D_fwd_norm = self.project_in_basis(D_fwd_at_coords, basis_vector=normals)
        D_adj_norm = self.project_in_basis(D_adj_at_coords, basis_vector=normals)

        E_fwd_perp1 = self.project_in_basis(E_fwd_at_coords, basis_vector=perps1)
        E_adj_perp1 = self.project_in_basis(E_adj_at_coords, basis_vector=perps1)

        E_fwd_perp2 = self.project_in_basis(E_fwd_at_coords, basis_vector=perps2)
        E_adj_perp2 = self.project_in_basis(E_adj_at_coords, basis_vector=perps2)

        # multiply forward and adjoint
        D_der_norm = D_fwd_norm * D_adj_norm
        E_der_perp1 = E_fwd_perp1 * E_adj_perp1
        E_der_perp2 = E_fwd_perp2 * E_adj_perp2

        # approximate permittivity in and out
        delta_eps_inv = self.delta_eps_inv(spatial_coords=spatial_coords)
        delta_eps = self.delta_eps(spatial_coords=spatial_coords)

        # put together VJP using D_normal and E_perp integration
        vjps = 0.0

        # perform D-normal integral
        contrib_D = -delta_eps_inv * D_der_norm
        vjps += contrib_D

        # perform E-perpendicular integrals
        for E_der in (E_der_perp1, E_der_perp2):
            contrib_E = E_der * delta_eps
            vjps += contrib_E

        return surface_mesh.areas * vjps

    def evaluate_eps(
        self,
        spatial_coords: np.ndarray,  # (N, 3)
        is_inside: bool,
    ) -> SpatialDataArray:
        """Evaluate permittivity without the structure at a set of points."""

        permittivity_array = self.eps_inf_structure if is_inside else self.eps_no_structure

        if permittivity_array is None:
            raise ValueError("Can't evaluate eps because the permittivity array is missing.")

        key = "key"
        eps_out = self.evaluate_flds_at(
            fld_dataset={key: permittivity_array},
            spatial_coords=spatial_coords,
        )[key]
        return eps_out.values

    @staticmethod
    def evaluate_flds_at(
        fld_dataset: dict[str, ScalarFieldDataArray],
        spatial_coords: np.ndarray,  # (N, 3)
    ) -> dict[str, ScalarFieldDataArray]:
        """Compute the value of an dict with keys Ex, Ey, Ez at a set of spatial locations."""

        xs, ys, zs = spatial_coords.T
        edge_index_dim = "edge_index"

        interp_kwargs = {}
        for dim, locations_dim in zip("xyz", (xs, ys, zs)):
            # filter out any infinity values to use LARGE_NUMBER and get 0 in the interp()
            locations_dim = np.nan_to_num(locations_dim, posinf=LARGE_NUMBER, neginf=-LARGE_NUMBER)
            interp_kwargs[dim] = xr.DataArray(locations_dim, dims=edge_index_dim)

        components = {}
        for fld_name, arr in fld_dataset.items():
            arr_interp = arr.interp(
                **interp_kwargs,
                assume_sorted=True,
                kwargs=dict(fill_value=None, bounds_error=False),
            )
            if "f" in arr_interp.coords:
                arr_interp = arr_interp.sum("f")

            components[fld_name] = arr_interp

        return components

    @staticmethod
    def project_in_basis(
        der_dataset: xr.Dataset,
        basis_vector: np.ndarray,
    ) -> xr.DataArray:
        """Project a derivative dataset along a supplied basis vector."""

        value = 0.0
        for coeffs, dim in zip(basis_vector.T, "xyz"):
            value += coeffs * der_dataset[f"E{dim}"]
        return value


# TODO: could we move this into a DataArray method?
def integrate_within_bounds(arr: xr.DataArray, dims: list[str], bounds: Bound) -> xr.DataArray:
    """integrate a data array within bounds, assumes bounds are [2, N] for N dims."""

    # order bounds with dimension first (N, 2)
    bounds = np.asarray(bounds).T
    all_coords = {}

    # loop over all dimensions
    for dim, (bmin, bmax) in zip(dims, bounds):
        bmin = get_static(bmin)
        bmax = get_static(bmax)

        coord_values = np.copy(arr.coords[dim].data)

        # reset all coordinates outside of bounds to the bounds, so that dL = 0 in integral
        np.clip(coord_values, bmin, bmax, out=coord_values)

        all_coords[dim] = coord_values

    _arr = arr.assign_coords(**all_coords)

    # uses trapezoidal rule
    # https://docs.xarray.dev/en/stable/generated/xarray.DataArray.integrate.html
    dims_integrate = [dim for dim in dims if len(_arr.coords[dim]) > 1]
    return _arr.integrate(coord=dims_integrate)


__all__ = [
    "integrate_within_bounds",
    "DerivativeInfo",
]



"""Stub for webapi"""

from __future__ import annotations

import json
from typing import Callable, List, Union

import pydantic.v1 as pd
from pydantic.v1 import BaseModel

from tidy3d.components.tcad.data.sim_data import HeatChargeSimulationData, HeatSimulationData
from tidy3d.components.tcad.simulation.heat import HeatSimulation
from tidy3d.components.tcad.simulation.heat_charge import HeatChargeSimulation

from ... import log
from ...components.base import _get_valid_extension
from ...components.data.monitor_data import ModeSolverData
from ...components.data.sim_data import SimulationData
from ...components.eme.data.sim_data import EMESimulationData
from ...components.eme.simulation import EMESimulation
from ...components.mode.data.sim_data import ModeSimulationData
from ...components.mode.simulation import ModeSimulation
from ...components.simulation import Simulation
from ...plugins.mode.mode_solver import ModeSolver
from ..core.file_util import (
    read_simulation_from_hdf5,
    read_simulation_from_hdf5_gz,
    read_simulation_from_json,
)
from ..core.stub import TaskStub, TaskStubData
from ..core.types import TaskType

SimulationType = Union[
    Simulation, HeatChargeSimulation, HeatSimulation, EMESimulation, ModeSolver, ModeSimulation
]
SimulationDataType = Union[
    SimulationData,
    HeatChargeSimulationData,
    HeatSimulationData,
    EMESimulationData,
    ModeSolverData,
    ModeSimulationData,
]


class Tidy3dStub(BaseModel, TaskStub):
    simulation: SimulationType = pd.Field(discriminator="type")

    @classmethod
    def from_file(cls, file_path: str) -> SimulationType:
        """Loads a Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        from .yaml, .json, or .hdf5 file.

        Parameters
        ----------
        file_path : str
            Full path to the .yaml or .json or .hdf5 file to load the
            Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`] from.

        Returns
        -------
        Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
            An instance of the component class calling ``load``.

        Example
        -------
        >>> simulation = Simulation.from_file(fname='folder/sim.json') # doctest: +SKIP
        """
        extension = _get_valid_extension(file_path)
        if extension == ".json":
            json_str = read_simulation_from_json(file_path)
        elif extension == ".hdf5":
            json_str = read_simulation_from_hdf5(file_path)
        elif extension == ".hdf5.gz":
            json_str = read_simulation_from_hdf5_gz(file_path)

        data = json.loads(json_str)
        type_ = data["type"]
        if "Simulation" == type_:
            sim = Simulation.from_file(file_path)
        elif "ModeSolver" == type_:
            sim = ModeSolver.from_file(file_path)
        elif "HeatSimulation" == type_:
            sim = HeatSimulation.from_file(file_path)
        elif "HeatChargeSimulation" == type_:
            sim = HeatChargeSimulation.from_file(file_path)
        elif "EMESimulation" == type_:
            sim = EMESimulation.from_file(file_path)
        elif "ModeSimulation" == type_:
            sim = ModeSimulation.from_file(file_path)

        return sim

    def to_file(
        self,
        file_path: str,
    ):
        """Exports Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`] instance to .yaml, .json,
        or .hdf5 file

        Parameters
        ----------
        file_path : str
            Full path to the .yaml or .json or .hdf5 file to save the :class:`Stub` to.

        Example
        -------
        >>> simulation.to_file(fname='folder/sim.json') # doctest: +SKIP
        """
        self.simulation.to_file(file_path)

    def to_hdf5_gz(self, fname: str, custom_encoders: List[Callable] = None) -> None:
        """Exports Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`] instance to .hdf5.gz file.

        Parameters
        ----------
        fname : str
            Full path to the .hdf5.gz file to save
            the Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`] to.
        custom_encoders : List[Callable]
            List of functions accepting (fname: str, group_path: str, value: Any) that take
            the ``value`` supplied and write it to the hdf5 ``fname`` at ``group_path``.

        Example
        -------
        >>> simulation.to_hdf5_gz(fname='folder/sim.hdf5.gz') # doctest: +SKIP
        """

        self.simulation.to_hdf5_gz(fname)

    def get_type(self) -> str:
        """Get simulation instance type.

        Returns
        -------
        :class:`TaskType`
            An instance Type of the component class calling ``load``.
        """
        if isinstance(self.simulation, Simulation):
            return TaskType.FDTD.name
        elif isinstance(self.simulation, ModeSolver):
            return TaskType.MODE_SOLVER.name
        elif isinstance(self.simulation, HeatSimulation):
            return TaskType.HEAT.name
        elif isinstance(self.simulation, HeatChargeSimulation):
            return TaskType.HEAT_CHARGE.name
        elif isinstance(self.simulation, EMESimulation):
            return TaskType.EME.name
        elif isinstance(self.simulation, ModeSimulation):
            return TaskType.MODE.name

    def validate_pre_upload(self, source_required) -> None:
        """Perform some pre-checks on instances of component"""
        if isinstance(self.simulation, Simulation):
            self.simulation.validate_pre_upload(source_required)
        elif isinstance(self.simulation, EMESimulation):
            self.simulation.validate_pre_upload()


class Tidy3dStubData(BaseModel, TaskStubData):
    """"""

    data: SimulationDataType

    @classmethod
    def from_file(cls, file_path: str) -> SimulationDataType:
        """Loads a Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
        from .yaml, .json, or .hdf5 file.

        Parameters
        ----------
        file_path : str
            Full path to the .yaml or .json or .hdf5 file to load the
            Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`] from.

        Returns
        -------
        Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
            An instance of the component class calling ``load``.
        """
        extension = _get_valid_extension(file_path)
        if extension == ".json":
            json_str = read_simulation_from_json(file_path)
        elif extension == ".hdf5":
            json_str = read_simulation_from_hdf5(file_path)
        elif extension == ".hdf5.gz":
            json_str = read_simulation_from_hdf5_gz(file_path)

        data = json.loads(json_str)
        type_ = data["type"]
        if "SimulationData" == type_:
            sim_data = SimulationData.from_file(file_path)
        elif "ModeSolverData" == type_:
            sim_data = ModeSolverData.from_file(file_path)
        elif "HeatSimulationData" == type_:
            sim_data = HeatSimulationData.from_file(file_path)
        elif "HeatChargeSimulationData" == type_:
            sim_data = HeatChargeSimulationData.from_file(file_path)
        elif "EMESimulationData" == type_:
            sim_data = EMESimulationData.from_file(file_path)
        elif "ModeSimulationData" == type_:
            sim_data = ModeSimulationData.from_file(file_path)

        return sim_data

    def to_file(self, file_path: str):
        """Exports Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`] instance
        to .yaml, .json, or .hdf5 file

        Parameters
        ----------
        file_path : str
            Full path to the .yaml or .json or .hdf5 file to save the
            Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`] to.

        Example
        -------
        >>> simulation.to_file(fname='folder/sim.json') # doctest: +SKIP
        """
        self.data.to_file(file_path)

    @classmethod
    def postprocess(cls, file_path: str) -> SimulationDataType:
        """Load .yaml, .json, or .hdf5 file to
        Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`] instance.

        Parameters
        ----------
        file_path : str
            Full path to the .yaml or .json or .hdf5 file to save the
            Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`] to.

        Returns
        -------
        Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
            An instance of the component class calling ``load``.
        """
        stub_data = Tidy3dStubData.from_file(file_path)

        check_log_msg = "For more information, check 'SimulationData.log' or use "
        check_log_msg += "'web.download_log(task_id)'."
        warned_about_warnings = False

        if isinstance(stub_data, SimulationData):
            final_decay_value = stub_data.final_decay_value
            shutoff_value = stub_data.simulation.shutoff
            if stub_data.diverged:
                log.warning("The simulation has diverged! " + check_log_msg)
                warned_about_warnings = True
            elif (shutoff_value != 0) and (final_decay_value > shutoff_value):
                log.warning(
                    f"Simulation final field decay value of {final_decay_value} is greater than "
                    f"the simulation shutoff threshold of {shutoff_value}. Consider running the "
                    "simulation again with a larger 'run_time' duration for more accurate results."
                )

        if (
            not isinstance(stub_data, (ModeSolverData, ModeSimulationData))
            and "WARNING" in stub_data.log
            and not warned_about_warnings
        ):
            log.warning("Warning messages were found in the solver log. " + check_log_msg)

        return stub_data



"""Defines a jax-compatible structure and its conversion to a gradient monitor."""

from __future__ import annotations

from typing import Dict, List, Union

import numpy as np
import pydantic.v1 as pd
from jax.tree_util import register_pytree_node_class

from ....components.data.monitor_data import FieldData, PermittivityData
from ....components.geometry.utils import GeometryType
from ....components.medium import MediumType
from ....components.monitor import FieldMonitor
from ....components.structure import Structure
from ....components.types import TYPE_TAG_STR, Bound
from ....constants import C_0
from .base import JaxObject
from .geometry import JAX_GEOMETRY_MAP, JaxBox, JaxGeometryType
from .medium import JAX_MEDIUM_MAP, JaxMediumType

GEO_MED_MAPPINGS = dict(geometry=JAX_GEOMETRY_MAP, medium=JAX_MEDIUM_MAP)


class AbstractJaxStructure(Structure, JaxObject):
    """A :class:`.Structure` registered with jax."""

    _tidy3d_class = Structure

    # which of "geometry" or "medium" is differentiable for this class
    _differentiable_fields = ()

    geometry: Union[JaxGeometryType, GeometryType]
    medium: Union[JaxMediumType, MediumType]

    @pd.validator("medium", always=True)
    def _check_2d_geometry(cls, val, values):
        """Override validator checking 2D geometry, which triggers unnecessarily for gradients."""
        return val

    def _validate_web_adjoint(self) -> None:
        """Run validators for this component, only if using ``tda.web.run()``."""
        if "geometry" in self._differentiable_fields:
            self.geometry._validate_web_adjoint()
        if "medium" in self._differentiable_fields:
            self.medium._validate_web_adjoint()

    @property
    def jax_fields(self):
        """The fields that are jax-traced for this class."""
        return dict(geometry=self.geometry, medium=self.medium)

    @property
    def exclude_fields(self):
        """Fields to exclude from the self dict."""
        return set(["type"] + list(self.jax_fields.keys()))

    def to_structure(self) -> Structure:
        """Convert :class:`.JaxStructure` instance to :class:`.Structure`"""
        self_dict = self.dict(exclude=self.exclude_fields)
        for key, component in self.jax_fields.items():
            if key in self._differentiable_fields:
                self_dict[key] = component.to_tidy3d()
            else:
                self_dict[key] = component
        return Structure.parse_obj(self_dict)

    @classmethod
    def from_structure(cls, structure: Structure) -> JaxStructure:
        """Convert :class:`.Structure` to :class:`.JaxStructure`."""

        struct_dict = structure.dict(exclude={"type"})

        jax_fields = dict(geometry=structure.geometry, medium=structure.medium)

        for key, component in jax_fields.items():
            if key in cls._differentiable_fields:
                type_map = GEO_MED_MAPPINGS[key]
                jax_type = type_map[type(component)]
                struct_dict[key] = jax_type.from_tidy3d(component)
            else:
                struct_dict[key] = component

        return cls.parse_obj(struct_dict)

    def make_grad_monitors(self, freqs: List[float], name: str) -> FieldMonitor:
        """Return gradient monitor associated with this object."""
        if "geometry" not in self._differentiable_fields:
            # make a fake JaxBox to be able to call .make_grad_monitors
            rmin, rmax = self.geometry.bounds
            geometry = JaxBox.from_bounds(rmin=rmin, rmax=rmax)
        else:
            geometry = self.geometry
        return geometry.make_grad_monitors(freqs=freqs, name=name)

    def _get_medium_params(
        self,
        grad_data_eps: PermittivityData,
    ) -> Dict[str, float]:
        """Compute params in the material of this structure."""
        freq_max = float(max(grad_data_eps.eps_xx.f))
        eps_in = self.medium.eps_model(frequency=freq_max)
        ref_ind = np.sqrt(np.max(np.real(eps_in)))
        ref_ind = max([1.0, abs(ref_ind)])
        wvl_free_space = C_0 / freq_max
        wvl_mat = wvl_free_space / ref_ind
        return dict(wvl_mat=wvl_mat, eps_in=eps_in)

    def geometry_vjp(
        self,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        eps_out: complex,
        num_proc: int = 1,
    ) -> JaxGeometryType:
        """Compute the VJP for the structure geometry."""

        medium_params = self._get_medium_params(grad_data_eps=grad_data_eps)

        return self.geometry.store_vjp(
            grad_data_fwd=grad_data_fwd,
            grad_data_adj=grad_data_adj,
            grad_data_eps=grad_data_eps,
            sim_bounds=sim_bounds,
            wvl_mat=medium_params["wvl_mat"],
            eps_out=eps_out,
            eps_in=medium_params["eps_in"],
            num_proc=num_proc,
        )

    def medium_vjp(
        self,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
    ) -> JaxMediumType:
        """Compute the VJP for the structure medium."""

        medium_params = self._get_medium_params(grad_data_eps=grad_data_eps)

        return self.medium.store_vjp(
            grad_data_fwd=grad_data_fwd,
            grad_data_adj=grad_data_adj,
            sim_bounds=sim_bounds,
            wvl_mat=medium_params["wvl_mat"],
            inside_fn=self.geometry.inside,
        )

    def store_vjp(
        self,
        # field_keys: List[Literal["medium", "geometry"]],
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        eps_out: complex,
        num_proc: int = 1,
    ) -> JaxStructure:
        """Returns the gradient of the structure parameters given forward and adjoint field data."""

        # return right away if field_keys are not present for some reason
        if not self._differentiable_fields:
            return self

        vjp_dict = {}

        # compute minimum wavelength in material (to use for determining integration points)
        if "geometry" in self._differentiable_fields:
            vjp_dict["geometry"] = self.geometry_vjp(
                grad_data_fwd=grad_data_fwd,
                grad_data_adj=grad_data_adj,
                grad_data_eps=grad_data_eps,
                sim_bounds=sim_bounds,
                eps_out=eps_out,
                num_proc=num_proc,
            )

        if "medium" in self._differentiable_fields:
            vjp_dict["medium"] = self.medium_vjp(
                grad_data_fwd=grad_data_fwd,
                grad_data_adj=grad_data_adj,
                grad_data_eps=grad_data_eps,
                sim_bounds=sim_bounds,
            )

        return self.updated_copy(**vjp_dict)


@register_pytree_node_class
class JaxStructure(AbstractJaxStructure, JaxObject):
    """A :class:`.Structure` registered with jax."""

    geometry: JaxGeometryType = pd.Field(
        ...,
        title="Geometry",
        description="Geometry of the structure, which is jax-compatible.",
        jax_field=True,
        discriminator=TYPE_TAG_STR,
    )

    medium: JaxMediumType = pd.Field(
        ...,
        title="Medium",
        description="Medium of the structure, which is jax-compatible.",
        jax_field=True,
        discriminator=TYPE_TAG_STR,
    )

    _differentiable_fields = ("medium", "geometry")


@register_pytree_node_class
class JaxStructureStaticMedium(AbstractJaxStructure, JaxObject):
    """A :class:`.Structure` registered with jax."""

    geometry: JaxGeometryType = pd.Field(
        ...,
        title="Geometry",
        description="Geometry of the structure, which is jax-compatible.",
        jax_field=True,
        discriminator=TYPE_TAG_STR,
    )

    medium: MediumType = pd.Field(
        ...,
        title="Medium",
        description="Regular ``tidy3d`` medium of the structure, non differentiable. "
        "Supports dispersive materials.",
        jax_field=False,
        discriminator=TYPE_TAG_STR,
    )

    _differentiable_fields = ("geometry",)


@register_pytree_node_class
class JaxStructureStaticGeometry(AbstractJaxStructure, JaxObject):
    """A :class:`.Structure` registered with jax."""

    geometry: GeometryType = pd.Field(
        ...,
        title="Geometry",
        description="Regular ``tidy3d`` geometry of the structure, non differentiable. "
        "Supports angled sidewalls and other complex geometries.",
        jax_field=False,
        discriminator=TYPE_TAG_STR,
    )

    medium: JaxMediumType = pd.Field(
        ...,
        title="Medium",
        description="Medium of the structure, which is jax-compatible.",
        jax_field=True,
        discriminator=TYPE_TAG_STR,
    )

    _differentiable_fields = ("medium",)


JaxStructureType = Union[JaxStructure, JaxStructureStaticMedium, JaxStructureStaticGeometry]



import scipy.ndimage
from autograd.extend import defvjp, primitive

gaussian_filter = primitive(scipy.ndimage.gaussian_filter)
defvjp(
    gaussian_filter,
    lambda ans, x, *args, **kwargs: lambda g: gaussian_filter(g, *args, **kwargs),
)



"""Compatibility layer for handling differences between package versions."""

try:
    from xarray.structure import alignment
except ImportError:
    from xarray.core import alignment

__all__ = ["alignment"]



"""
Our DC sources ultimately need to follow this standard form if we want to enable full electrical integration.

```
11.3.2 .DC: DC Transfer Function

General form:

    .dc srcnam vstart vstop vincr [src2 start2 stop2 incr2]

Examples:

    .dc VIN 0.25 5.0 0.25
    .dc VDS 0 10 .5 VGS 0 5 1
    .dc VCE 0 10 .25 IB 0 10u 1u
    .dc RLoad 1k 2k 100
    .dc TEMP -15 75 5
```

"""

from typing import Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.constants import AMP, VOLT


class DCVoltageSource(Tidy3dBaseModel):
    """
    DC voltage source in volts.

    Notes
    -----

        This voltage refers to potential above the equivalent simulation ground. Currently, electrical ports
        are not defined.

    Examples
    --------
    >>> import tidy3d as td
    >>> voltages = [-0.5, 0, 1, 2, 3, 4]
    >>> voltage_source = td.DCVoltageSource(voltage=voltages)
    """

    name: Optional[str]
    voltage: Union[pd.FiniteFloat, list[pd.FiniteFloat]] = pd.Field(
        title="Voltage",
        description="DC voltage usually used as source in 'VoltageBC' boundary conditions.",
    )
    units: str = VOLT


class DCCurrentSource(Tidy3dBaseModel):
    """
    DC current source in amperes.

    Example
    -------
    >>> import tidy3d as td
    >>> current_source = td.DCCurrentSource(current=0.4)
    """

    name: Optional[str]
    current: pd.FiniteFloat = pd.Field(
        title="Current",
        description="DC current usually used as source in 'CurrentBC' boundary conditions.",
    )
    units: str = AMP



"""Imports for adjoint plugin."""

# import the jax version of tidy3d components
try:
    import jax

    jax.config.update("jax_enable_x64", True)
except ImportError as e:
    raise ImportError(
        "The 'jax' package is required for adjoint plugin. We were not able to import it. "
        "To get the appropriate packages for your system, install tidy3d using '[jax]' option, "
        "for example: $pip install 'tidy3d[jax]'."
    ) from e

from .components.data.data_array import JaxDataArray
from .components.data.dataset import JaxPermittivityDataset
from .components.data.monitor_data import JaxModeData
from .components.data.sim_data import JaxSimulationData
from .components.geometry import JaxBox, JaxComplexPolySlab, JaxGeometryGroup, JaxPolySlab
from .components.medium import JaxAnisotropicMedium, JaxCustomMedium, JaxMedium
from .components.simulation import JaxSimulation
from .components.structure import (
    JaxStructure,
    JaxStructureStaticGeometry,
    JaxStructureStaticMedium,
)
from .web import run, run_async

__all__ = [
    "JaxBox",
    "JaxPolySlab",
    "JaxComplexPolySlab",
    "JaxGeometryGroup",
    "JaxMedium",
    "JaxAnisotropicMedium",
    "JaxCustomMedium",
    "JaxStructure",
    "JaxStructureStaticMedium",
    "JaxStructureStaticGeometry",
    "JaxSimulation",
    "JaxSimulationData",
    "JaxModeData",
    "JaxPermittivityDataset",
    "JaxDataArray",
    "run",
    "run_async",
]



"""File containing classes required for the setup of a DEVSIM case."""

from tidy3d.components.tcad.bandgap import SlotboomBandGapNarrowing
from tidy3d.components.tcad.boundary.charge import CurrentBC, InsulatingBC, VoltageBC
from tidy3d.components.tcad.boundary.heat import ConvectionBC, HeatFluxBC, TemperatureBC
from tidy3d.components.tcad.generation_recombination import (
    AugerRecombination,
    RadiativeRecombination,
    ShockleyReedHallRecombination,
)
from tidy3d.components.tcad.mobility import CaugheyThomasMobility, ConstantMobilityModel
from tidy3d.components.tcad.monitors.charge import (
    SteadyCapacitanceMonitor,
    SteadyFreeCarrierMonitor,
    SteadyPotentialMonitor,
)
from tidy3d.components.tcad.monitors.heat import TemperatureMonitor
from tidy3d.components.tcad.source.coupled import HeatFromElectricSource
from tidy3d.components.tcad.source.heat import HeatSource, UniformHeatSource
from tidy3d.components.types import Union

MobilityModelType = Union[CaugheyThomasMobility, ConstantMobilityModel]
RecombinationModelType = Union[
    AugerRecombination, RadiativeRecombination, ShockleyReedHallRecombination
]
BandGapNarrowingModelType = Union[SlotboomBandGapNarrowing]

# types of monitors that are accepted by heat simulation
HeatChargeMonitorType = Union[
    TemperatureMonitor,
    SteadyPotentialMonitor,
    SteadyFreeCarrierMonitor,
    SteadyCapacitanceMonitor,
]
HeatChargeSourceType = Union[HeatSource, HeatFromElectricSource, UniformHeatSource]
HeatChargeBCType = Union[
    TemperatureBC, HeatFluxBC, ConvectionBC, VoltageBC, CurrentBC, InsulatingBC
]



"""Monitor level data, store the DataArrays associated with a single heat-charge monitor."""

from __future__ import annotations

from typing import Dict, Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import skip_if_fields_missing
from tidy3d.components.data.data_array import (
    DataArray,
    IndexedDataArray,
    SpatialDataArray,
)
from tidy3d.components.data.utils import TetrahedralGridDataset, TriangularGridDataset
from tidy3d.components.tcad.data.monitor_data.abstract import HeatChargeMonitorData
from tidy3d.components.tcad.monitors.heat import (
    TemperatureMonitor,
)
from tidy3d.components.types import annotate_type
from tidy3d.constants import KELVIN
from tidy3d.log import log

FieldDataset = Union[
    SpatialDataArray, annotate_type(Union[TriangularGridDataset, TetrahedralGridDataset])
]
UnstructuredFieldType = Union[TriangularGridDataset, TetrahedralGridDataset]


class TemperatureData(HeatChargeMonitorData):
    """Data associated with a :class:`TemperatureMonitor`: spatial temperature field.

    Example
    -------
    >>> from tidy3d import TemperatureMonitor, SpatialDataArray
    >>> import numpy as np
    >>> temp_data = SpatialDataArray(
    ...     np.ones((2, 3, 4)), coords={"x": [0, 1], "y": [0, 1, 2], "z": [0, 1, 2, 3]}
    ... )
    >>> temp_mnt = TemperatureMonitor(size=(1, 2, 3), name="temperature")
    >>> temp_mnt_data = TemperatureData(
    ...     monitor=temp_mnt, temperature=temp_data, symmetry=(0, 1, 0), symmetry_center=(0, 0, 0)
    ... )
    >>> temp_mnt_data_expanded = temp_mnt_data.symmetry_expanded_copy
    """

    monitor: TemperatureMonitor = pd.Field(
        ..., title="Monitor", description="Temperature monitor associated with the data."
    )

    temperature: Optional[FieldDataset] = pd.Field(
        ...,
        title="Temperature",
        description="Spatial temperature field.",
        units=KELVIN,
    )

    @property
    def field_components(self) -> Dict[str, DataArray]:
        """Maps the field components to their associated data."""
        return dict(temperature=self.temperature)

    @pd.validator("temperature", always=True)
    @skip_if_fields_missing(["monitor"])
    def warn_no_data(cls, val, values):
        """Warn if no data provided."""

        mnt = values.get("monitor")

        if val is None:
            log.warning(
                f"No data is available for monitor '{mnt.name}'. This is typically caused by "
                "monitor not intersecting any solid medium."
            )

        return val

    @pd.validator("temperature", always=True)
    @skip_if_fields_missing(["monitor"])
    def check_correct_data_type(cls, val, values):
        """Issue error if incorrect data type is used"""

        mnt = values.get("monitor")

        if isinstance(val, TetrahedralGridDataset) or isinstance(val, TriangularGridDataset):
            if not isinstance(val.values, IndexedDataArray):
                raise ValueError(
                    f"Monitor {mnt} of type 'TemperatureMonitor' cannot be associated with data arrays "
                    "of type 'IndexVoltageDataArray'."
                )

        return val

    def field_name(self, val: str = "") -> str:
        """Gets the name of the fields to be plot."""
        if val == "abs^2":
            return "|T|², K²"
        else:
            return "T, K"

    @property
    def symmetry_expanded_copy(self) -> TemperatureData:
        """Return copy of self with symmetry applied."""

        new_temp = self._symmetry_expanded_copy(property=self.temperature)
        return self.updated_copy(temperature=new_temp, symmetry=(0, 0, 0))



"""Mode solver for propagating EM modes."""

from typing import Tuple

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sp
import scipy.sparse.linalg as spl

from ...constants import C_0, ETA_0, fp_eps, pec_val
from ..base import Tidy3dBaseModel
from ..types import EpsSpecType, ModeSolverType, Numpy
from .derivatives import create_d_matrices as d_mats
from .derivatives import create_s_matrices as s_mats
from .transforms import angled_transform, radial_transform

# Consider vec to be complex if norm(vec.imag)/norm(vec) > TOL_COMPLEX
TOL_COMPLEX = 1e-10
# Tolerance for eigs
TOL_EIGS = fp_eps
# Tolerance for deciding on the matrix to be diagonal or tensorial
TOL_TENSORIAL = 1e-6
# shift target neff by this value, both rel and abs, whichever results in larger shift
TARGET_SHIFT = 10 * fp_eps
# Preconditioner: "Jacobi" or "Material" based
PRECONDITIONER = "Material"
# Good conductor permittivity cut-off value. Let it be as large as possible so long as not causing overflow in
# double precision. This value is very heuristic.
GOOD_CONDUCTOR_CUT_OFF = 1e70
# Consider a material to be good conductor if |ep| (or |mu|) > GOOD_CONDUCTOR_THRESHOLD * |pec_val|
GOOD_CONDUCTOR_THRESHOLD = 0.9


class EigSolver(Tidy3dBaseModel):
    """Interface for computing eigenvalues given permittivity and mode spec.
    It's a collection of static methods.
    """

    @classmethod
    def compute_modes(
        cls,
        eps_cross,
        coords,
        freq,
        mode_spec,
        precision,
        mu_cross=None,
        split_curl_scaling=None,
        symmetry=(0, 0),
        direction="+",
        solver_basis_fields=None,
        plane_center: tuple[float, float] = None,
    ) -> Tuple[Numpy, Numpy, EpsSpecType]:
        """
        Solve for the modes of a waveguide cross-section.

        Parameters
        ----------
        eps_cross : array_like or tuple of array_like
            Either a single 2D array defining the relative permittivity in the cross-section,
            or nine 2D arrays defining the permittivity at the Ex, Ey, and Ez locations
            of the Yee grid in the order xx, xy, xz, yx, yy, yz, zx, zy, zz.
        coords : List[Numpy]
            Two 1D arrays with each with size one larger than the corresponding axis of
            ``eps_cross``.
            Defines a (potentially non-uniform) Cartesian grid on which the modes are computed.
        freq : float
            (Hertz) Frequency at which the eigenmodes are computed.
        mode_spec : ModeSpec
            ``ModeSpec`` object containing specifications of the mode solver.
        precision : Literal["single", "double"]
            Apply "single" or "double" precision in mode solver.
        mu_cross : array_like or tuple of array_like
            Either a single 2D array defining the relative permeability in the cross-section,
            or nine 2D arrays defining the permeability at the Hx, Hy, and Hz locations
            of the Yee grid in the order xx, xy, xz, yx, yy, yz, zx, zy, zz.
            Set to 1 if `None`.
        split_curl_scaling : tuple of array_like
            Split curl coefficient to Curl E. Three 2D arrays defining the scaling factor
            at the Ex, Ey, and Ez locations of the Yee grid in the order xx, yy, zz.
            Following Benkler's approach, we formulate it as the following:
            1) during mode solver: eps_cross -> eps_corss / scaling, so eigenvector is E * scaling
            2) in postprocessing: apply scaling^-1 to eigenvector to obtain E
        direction : Union["+", "-"]
            Direction of mode propagation.
        solver_basis_fields
            If provided, solve for modes in this basis.
        plane_center
            The center of the mode plane along the tangential axes of the global simulation. Used
            in case of bend modes to offset the coordinates correctly w.r.t. the bend radius, which
            is assumed to refer to the distance from the bend center to the mode plane center.

        Returns
        -------
        Tuple[Numpy, Numpy, str]
            The first array gives the E and H fields for all modes, the second one gives the complex
            effective index. The last variable describes permittivity characterization on the mode
            solver's plane ("diagonal", "tensorial_real", or "tensorial_complex").
        """

        num_modes = mode_spec.num_modes
        bend_radius = mode_spec.bend_radius
        bend_axis = mode_spec.bend_axis
        angle_theta = mode_spec.angle_theta
        angle_phi = mode_spec.angle_phi
        omega = 2 * np.pi * freq
        k0 = omega / C_0
        enable_incidence_matrices = False  # Experimental feature, always off for now

        eps_formated = cls.format_medium_data(eps_cross)
        eps_xx, eps_xy, eps_xz, eps_yx, eps_yy, eps_yz, eps_zx, eps_zy, eps_zz = eps_formated

        mu_formated = None
        if mu_cross is not None:
            mu_formated = cls.format_medium_data(mu_cross)

        Nxy = eps_xx.shape
        Nx, Ny = Nxy
        N = eps_xx.size

        if len(coords[0]) != Nx + 1 or len(coords[1]) != Ny + 1:
            raise ValueError("Mismatch between 'coords' and 'esp_cross' shapes.")
        new_coords = [np.copy(c) for c in coords]

        """We work with full tensorial epsilon in mu to handle the most general cases that can
        be introduced by coordinate transformations. In the solver, we distinguish the case when
        these tensors are still diagonal, in which case the matrix for diagonalization has shape
        (2N, 2N), and the full tensorial case, in which case it has shape (4N, 4N)."""
        eps_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        mu_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        identity_tensor = np.zeros((3, 3, N), dtype=np.complex128)
        for row, eps_row in enumerate(
            [[eps_xx, eps_xy, eps_xz], [eps_yx, eps_yy, eps_yz], [eps_zx, eps_zy, eps_zz]]
        ):
            identity_tensor[row, row, :] = 1.0
            for col, eps in enumerate(eps_row):
                if split_curl_scaling is not None and col == row:
                    outside_pec = ~np.isclose(split_curl_scaling[col], 0)
                    eps[outside_pec] /= split_curl_scaling[col][outside_pec]

                eps_tensor[row, col, :] = eps.ravel()

        if mu_formated is not None:
            mu_xx, mu_xy, mu_xz, mu_yx, mu_yy, mu_yz, mu_zx, mu_zy, mu_zz = mu_formated
            for row, mu_row in enumerate(
                [[mu_xx, mu_xy, mu_xz], [mu_yx, mu_yy, mu_yz], [mu_zx, mu_zy, mu_zz]]
            ):
                for col, mu in enumerate(mu_row):
                    mu_tensor[row, col, :] = mu.ravel()
        else:
            mu_tensor = np.copy(identity_tensor)

        # Get Jacobian of all coordinate transformations. Initialize as identity (same as mu so far)
        jac_e = np.real(np.copy(identity_tensor))
        jac_h = np.real(np.copy(identity_tensor))

        if np.abs(angle_theta) > 0:
            new_coords, jac_e, jac_h = angled_transform(new_coords, angle_theta, angle_phi)

        if bend_radius is not None:
            if plane_center is None:
                raise ValueError(
                    "When using a nonzero 'bend_radius', the 'plane_center' must also "
                    "be provided to correctly offset the coordinates."
                )
            new_coords, jac_e_tmp, jac_h_tmp = radial_transform(
                new_coords, bend_radius, bend_axis, plane_center
            )
            jac_e = np.einsum("ij...,jp...->ip...", jac_e_tmp, jac_e)
            jac_h = np.einsum("ij...,jp...->ip...", jac_h_tmp, jac_h)

        """We also need to keep track of the transformation of the k-vector. This is
        the eigenvalue of the momentum operator assuming some sort of translational invariance and is
        different from just the transformation of the derivative operator. For example, in a bent
        waveguide, there is strictly speaking no k-vector in the original coordinates as the system
        is not translationally invariant there. However, if we define kz = R k_phi, then the
        effective index approaches that for a straight-waveguide in the limit of infinite radius.
        Since we use w = R phi in the radial_transform, there is nothing else needed in the k transform.
        For the angled_transform, the transformation between k-vectors follows from writing the field as
        E' exp(i k_p w) in transformed coordinates, and identifying this with
        E exp(i k_x x + i k_y y + i k_z z) in the original ones."""
        kxy = np.cos(angle_theta) ** 2
        kz = np.cos(angle_theta) * np.sin(angle_theta)
        kp_to_k = np.array([kxy * np.sin(angle_phi), kxy * np.cos(angle_phi), kz])

        # Transform epsilon and mu
        jac_e_det = np.linalg.det(np.moveaxis(jac_e, [0, 1], [-2, -1]))
        jac_h_det = np.linalg.det(np.moveaxis(jac_h, [0, 1], [-2, -1]))
        eps_tensor = np.einsum("ij...,jp...->ip...", jac_e, eps_tensor)  # J.dot(eps)
        eps_tensor = np.einsum("ij...,pj...->ip...", eps_tensor, jac_e)  # (J.dot(eps)).dot(J.T)
        eps_tensor /= jac_e_det
        mu_tensor = np.einsum("ij...,jp...->ip...", jac_h, mu_tensor)
        mu_tensor = np.einsum("ij...,pj...->ip...", mu_tensor, jac_h)
        mu_tensor /= jac_h_det

        # # Uncomment block to force eps and mu to be translationally invariant into the PML.
        # # This may be important for bends as the jacobian transformation breaks the invariance, but
        # # tests reveal that it has little effect.
        # eps_tensor = cls.make_pml_invariant(Nxy, eps_tensor, mode_spec.num_pml)
        # mu_tensor = cls.make_pml_invariant(Nxy, mu_tensor, mode_spec.num_pml)

        """ Boundaries are imposed through the derivative matrices. The forward derivative matrices
        always impose PEC boundary at the xmax and ymax interfaces, and on the xmin and ymin
        interfaces unless PMC symmetry is present. If so, the PMC boundary is imposed through the
        backward derivative matrices."""
        dmin_pmc = [sym == 1 for sym in symmetry]

        # Primal grid steps for E-field derivatives
        dl_f = [new_cs[1:] - new_cs[:-1] for new_cs in new_coords]
        # Dual grid steps for H-field derivatives
        dl_tmp = [(dl[:-1] + dl[1:]) / 2 for dl in dl_f]
        dl_b = [np.hstack((d1[0], d2)) for d1, d2 in zip(dl_f, dl_tmp)]
        dls = (dl_f, dl_b)

        # Derivative matrices with PEC boundaries by default and optional PMC at the near end
        der_mats_tmp = d_mats(Nxy, dls, dmin_pmc)

        # PML matrices; do not impose PML on the bottom when symmetry present
        dmin_pml = np.array(symmetry) == 0
        pml_mats = s_mats(omega, Nxy, mode_spec.num_pml, dls, eps_tensor, mu_tensor, dmin_pml)

        # Add the PML on top of the derivatives; normalize by k0 to match the EM-possible notation
        der_mats = [Smat.dot(Dmat) / k0 for Smat, Dmat in zip(pml_mats, der_mats_tmp)]

        # Determine initial guess value for the solver in transformed coordinates
        if mode_spec.target_neff is None:
            eps_physical = np.array(eps_cross)
            eps_physical = eps_physical[np.abs(eps_physical) < np.abs(pec_val)]
            n_max = np.sqrt(np.max(np.abs(eps_physical)))
            target = n_max
        else:
            target = mode_spec.target_neff
        target_neff_p = target / np.linalg.norm(kp_to_k)

        # shift target_neff slightly to avoid cases where the shiftted matrix is exactly singular
        if abs(TARGET_SHIFT) > abs(target_neff_p * TARGET_SHIFT):
            target_neff_p += TARGET_SHIFT
        else:
            target_neff_p *= 1 + TARGET_SHIFT

        # preprocess solver_basis_fields
        basis_E = None
        if solver_basis_fields is not None:
            basis_E = solver_basis_fields[:3, ...]
            try:
                basis_E = basis_E.reshape((3, Nx * Ny, num_modes))
            except ValueError:
                raise ValueError(
                    "Shape mismatch between 'basis_fields' and requested mode data. "
                    "Make sure the mode solvers are set up the same, and that the "
                    "basis mode solver data has 'colocate=False'."
                )
            if split_curl_scaling is not None:
                basis_E = cls.split_curl_field_postprocess_inverse(split_curl_scaling, basis_E)
            jac_e_inv = np.moveaxis(
                np.linalg.inv(np.moveaxis(jac_e, [0, 1], [-2, -1])), [-2, -1], [0, 1]
            )
            basis_E = np.sum(jac_e_inv[..., None] * basis_E[:, None, ...], axis=0)

        # Solve for the modes
        E, H, neff, keff, eps_spec = cls.solver_em(
            Nx,
            Ny,
            eps_tensor,
            mu_tensor,
            der_mats,
            num_modes,
            target_neff_p,
            precision,
            direction,
            enable_incidence_matrices,
            basis_E=basis_E,
        )

        # Transform back to original axes, E = J^T E'
        E = np.sum(jac_e[..., None] * E[:, None, ...], axis=0)
        if split_curl_scaling is not None:
            E = cls.split_curl_field_postprocess(split_curl_scaling, E)
        E = E.reshape((3, Nx, Ny, 1, num_modes))
        H = np.sum(jac_h[..., None] * H[:, None, ...], axis=0)
        H = H.reshape((3, Nx, Ny, 1, num_modes))
        fields = np.stack((E, H), axis=0)

        neff = neff * np.linalg.norm(kp_to_k)
        keff = keff * np.linalg.norm(kp_to_k)

        if mode_spec.precision == "single":
            # Recast to single precision which may have changed due to earlier manipulations
            fields = fields.astype(np.complex64)

        return fields, neff + 1j * keff, eps_spec

    @classmethod
    def solver_em(
        cls,
        Nx,
        Ny,
        eps_tensor,
        mu_tensor,
        der_mats,
        num_modes,
        neff_guess,
        mat_precision,
        direction,
        enable_incidence_matrices,
        basis_E,
    ):
        """Solve for the electromagnetic modes of a system defined by in-plane permittivity and
        permeability and assuming translational invariance in the normal direction.

        Parameters
        ----------
        Nx : int
            Number of grids along x-direction.
        Ny : int
            Number of grids along y-direction.
        eps_tensor : np.ndarray
            Shape (3, 3, N), the permittivity tensor at every point in the plane.
        mu_tensor : np.ndarray
            Shape (3, 3, N), the permittivity tensor at every point in the plane.
        der_mats : List[scipy.sparse.csr_matrix]
            The sparce derivative matrices dxf, dxb, dyf, dyb, including the PML.
        num_modes : int
            Number of modes to solve for.
        neff_guess : float
            Initial guess for the effective index.
        mat_precision : Union['single', 'double']
            Single or double-point precision in eigensolver.
        direction : Union["+", "-"]
            Direction of mode propagation.
        basis_E: np.ndarray
            Basis for mode solving.

        Returns
        -------
        E : np.ndarray
            Electric field of the eigenmodes, shape (3, N, num_modes).
        H : np.ndarray
            Magnetic field of the eigenmodes, shape (3, N, num_modes).
        neff : np.ndarray
            Real part of the effective index, shape (num_modes, ).
        keff : np.ndarray
            Imaginary part of the effective index, shape (num_modes, ).
        eps_spec : Union["diagonal", "tensorial_real", "tensorial_complex"]
            Permittivity characterization on the mode solver's plane.
        """

        # In the matrices P and Q below, they contain terms ``epsilon_parallel`` or
        # ``mu_parallel``, and also a term proportional to 1/(k0 * dl)**2. To make sure
        # that permittivity of good conductor is visible in low-frequency/high resolution, pec_val should be
        # scaled by a factor max(1, max(1/k0 dl) **2).
        pec_scaling = max(1, max([np.max(abs(f)) for f in der_mats]) ** 2)
        pec_scaled_val = min(GOOD_CONDUCTOR_CUT_OFF, pec_scaling * abs(pec_val))

        # use a high-conductivity model for locations associated with a good conductor
        def conductivity_model_for_good_conductor(
            eps, threshold=GOOD_CONDUCTOR_THRESHOLD * pec_val
        ):
            """Entries associated with 'eps' are converted to a high-conductivity model."""
            eps = eps.astype(complex)
            eps[np.abs(eps) >= abs(threshold)] = 1 + 1j * pec_scaled_val
            return eps

        eps_tensor = conductivity_model_for_good_conductor(eps_tensor)
        mu_tensor = conductivity_model_for_good_conductor(mu_tensor)

        # Determine if ``eps`` and ``mu`` are diagonal or tensorial
        off_diagonals = (np.ones((3, 3)) - np.eye(3)).astype(bool)
        eps_offd = np.abs(eps_tensor[off_diagonals])
        mu_offd = np.abs(mu_tensor[off_diagonals])
        is_tensorial = np.any(eps_offd > TOL_TENSORIAL) or np.any(mu_offd > TOL_TENSORIAL)

        # initial vector for eigensolver in correct data type
        vec_init = cls.set_initial_vec(Nx, Ny, is_tensorial=is_tensorial)

        # call solver
        kwargs = {
            "eps": eps_tensor,
            "mu": mu_tensor,
            "der_mats": der_mats,
            "num_modes": num_modes,
            "neff_guess": neff_guess,
            "vec_init": vec_init,
            "mat_precision": mat_precision,
        }

        is_eps_complex = cls.isinstance_complex(eps_tensor)

        if basis_E is not None and is_tensorial:
            raise RuntimeError(
                "Tensorial eps not yet supported in relative mode solver "
                "(with basis fields provided)."
            )

        if not is_tensorial:
            eps_spec = "diagonal"
            E, H, neff, keff = cls.solver_diagonal(
                **kwargs,
                enable_incidence_matrices=enable_incidence_matrices,
                basis_E=basis_E,
            )
            if direction == "-":
                H[0] *= -1
                H[1] *= -1
                E[2] *= -1

        elif not is_eps_complex:
            eps_spec = "tensorial_real"
            E, H, neff, keff = cls.solver_tensorial(**kwargs, direction="+")
            if direction == "-":
                E = np.conj(E)
                H = -np.conj(H)

        else:
            eps_spec = "tensorial_complex"
            E, H, neff, keff = cls.solver_tensorial(**kwargs, direction=direction)

        return E, H, neff, keff, eps_spec

    @classmethod
    def matrix_data_type(cls, eps, mu, der_mats, mat_precision, is_tensorial):
        """Determine data type that should be used for the matrix for diagonalization."""
        mat_dtype = np.float32
        # In tensorial case, even though the matrix can be real, the
        # expected eigenvalue is purely imaginary. So for now we enforce
        # the matrix to be complex type so that it will look for the right eigenvalues.
        if is_tensorial:
            mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
        else:
            # 1) check if complex or not
            complex_solver = (
                cls.isinstance_complex(eps)
                or cls.isinstance_complex(mu)
                or np.any([cls.isinstance_complex(f) for f in der_mats])
            )
            # 2) determine precision
            if complex_solver:
                mat_dtype = np.complex128 if mat_precision == "double" else np.complex64
            else:
                if mat_precision == "double":
                    mat_dtype = np.float64

        return mat_dtype

    @classmethod
    def trim_small_values(cls, mat: sp.csr_matrix, tol: float) -> sp.csr_matrix:
        """Eliminate elements of matrix ``mat`` for which ``abs(element) / abs(max_element) < tol``,
        or ``np.abs(mat_data) < tol``. This operates in-place on mat so there is no return.
        """
        max_element = np.amax(np.abs(mat))
        mat.data *= np.logical_or(np.abs(mat.data) / max_element > tol, np.abs(mat.data) > tol)
        mat.eliminate_zeros()

    @classmethod
    def solver_diagonal(
        cls,
        eps,
        mu,
        der_mats,
        num_modes,
        neff_guess,
        vec_init,
        mat_precision,
        enable_incidence_matrices,
        basis_E,
    ):
        """EM eigenmode solver assuming ``eps`` and ``mu`` are diagonal everywhere."""

        # code associated with these options is included below in case it's useful in the future
        enable_preconditioner = False
        analyze_conditioning = False

        def incidence_matrix_for_pec(eps_vec, threshold=0.9 * np.abs(pec_val)):
            """Incidence matrix indicating non-PEC entries associated with 'eps_vec'."""
            nnz = eps_vec[np.abs(eps_vec) < threshold]
            eps_nz = eps_vec.copy()
            eps_nz[np.abs(eps_vec) >= threshold] = 0
            rows = np.arange(0, len(nnz))
            cols = np.argwhere(eps_nz).flatten()
            dnz = sp.csr_matrix(([1] * len(nnz), (rows, cols)), shape=(len(rows), len(eps_vec)))
            return dnz

        mode_solver_type = "diagonal"
        N = eps.shape[-1]

        # Unpack eps, mu and derivatives
        eps_xx = eps[0, 0, :]
        eps_yy = eps[1, 1, :]
        eps_zz = eps[2, 2, :]
        mu_xx = mu[0, 0, :]
        mu_yy = mu[1, 1, :]
        mu_zz = mu[2, 2, :]
        dxf, dxb, dyf, dyb = der_mats

        if any(
            cls.mode_plane_contain_good_conductor(i)
            for i in [eps_xx, eps_yy, eps_zz, mu_xx, mu_yy, mu_zz]
        ):
            enable_preconditioner = True

        # Compute the matrix for diagonalization
        inv_eps_zz = sp.spdiags(1 / eps_zz, [0], N, N)
        inv_mu_zz = sp.spdiags(1 / mu_zz, [0], N, N)

        if enable_incidence_matrices:
            dnz_xx, dnz_yy, dnz_zz = (incidence_matrix_for_pec(i) for i in [eps_xx, eps_yy, eps_zz])
            dnz = sp.block_diag((dnz_xx, dnz_yy), format="csr")
            inv_eps_zz = (dnz_zz.T) * dnz_zz * inv_eps_zz * (dnz_zz.T) * dnz_zz

        # P = p_mu + p_partial
        # Q = q_ep + q_partial
        # Note that p_partial @ q_partial = 0, so that PQ = p_mu @ Q + p_partial @ q_ep
        p_mu = sp.bmat(
            [[None, sp.spdiags(mu_yy, [0], N, N)], [sp.spdiags(-mu_xx, [0], N, N), None]]
        )
        p_partial = sp.bmat(
            [
                [-dxf.dot(inv_eps_zz).dot(dyb), dxf.dot(inv_eps_zz).dot(dxb)],
                [-dyf.dot(inv_eps_zz).dot(dyb), dyf.dot(inv_eps_zz).dot(dxb)],
            ]
        )
        q_ep = sp.bmat(
            [[None, sp.spdiags(eps_yy, [0], N, N)], [sp.spdiags(-eps_xx, [0], N, N), None]]
        )
        q_partial = sp.bmat(
            [
                [-dxb.dot(inv_mu_zz).dot(dyf), dxb.dot(inv_mu_zz).dot(dxf)],
                [-dyb.dot(inv_mu_zz).dot(dyf), dyb.dot(inv_mu_zz).dot(dxf)],
            ]
        )

        # pmat = p_mu + p_partial  # no need to assemble pmat, as it is not used anywhere
        qmat = q_ep + q_partial
        mat = p_mu @ qmat + p_partial @ q_ep

        # Cast matrix to target data type
        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=False)
        mat = cls.type_conversion(mat, mat_dtype)

        # Casting starting vector to target data type
        vec_init = cls.type_conversion(vec_init, mat_dtype)

        # Starting eigenvalue guess in target data type
        eig_guess = cls.type_conversion(np.array([-(neff_guess**2)]), mat_dtype)[0]

        if enable_incidence_matrices:
            mat = dnz * mat * dnz.T
            vec_init = dnz * vec_init

        # Denote the original eigenvalue problem as Ax = lambda x,
        # with left and right preconditioners, we solve for the following eigenvalue problem,
        # L A R y = lambda LR y, where x = R y
        precon_left = None
        precon_right = None
        generalized_M = None  # matrix in the generalized eigenvalue problem
        if enable_preconditioner:
            if PRECONDITIONER == "Jacobi":
                precon_right = sp.diags(1 / mat.diagonal())

            elif PRECONDITIONER == "Material":

                def conditional_inverted_vec(eps_vec, threshold=1):
                    """Returns a diagonal sparse matrix whose i-th element in the diagonal
                    is |eps_i|^-1 if |eps_i|>threshold, and |eps_i| otherwise.
                    """
                    abs_vec = np.abs(eps_vec)
                    return sp.spdiags(
                        np.where(abs_vec > threshold, 1.0 / abs_vec, abs_vec), [0], N, N
                    )

                precon_left = sp.bmat(
                    [
                        [conditional_inverted_vec(mu_yy), None],
                        [None, conditional_inverted_vec(mu_xx)],
                    ]
                )
                precon_right = sp.bmat(
                    [
                        [conditional_inverted_vec(eps_xx), None],
                        [None, conditional_inverted_vec(eps_yy)],
                    ]
                )
            generalized_M = precon_right
            mat = mat @ precon_right
            if precon_left is not None:
                generalized_M = precon_left @ generalized_M
                mat = precon_left @ mat

        # Trim small values in single precision case
        if mat_precision == "single":
            cls.trim_small_values(mat, tol=fp_eps)

        if analyze_conditioning:
            aca = mat.conjugate().T * mat
            aac = mat * mat.conjugate().T
            diff = aca - aac
            print(
                f"inf-norm: A*A: {spl.norm(aca, ord=np.inf)}, AA*: {spl.norm(aac, ord=np.inf)}, nonnormality: {spl.norm(diff, ord=np.inf)}, relative nonnormality: {spl.norm(diff, ord=np.inf)/spl.norm(aca, ord=np.inf)}"
            )
            print(
                f"fro-norm: A*A: {spl.norm(aca, ord='fro')}, AA*: {spl.norm(aac, ord='fro')}, nonnormality: {spl.norm(diff, ord='fro')}, relative nonnormality: {spl.norm(diff, ord='fro')/spl.norm(aca, ord='fro')}"
            )

        # preprocess basis modes
        basis_vecs = None
        if basis_E is not None:
            basis_Ex = basis_E[0, ...]
            basis_Ey = basis_E[1, ...]
            basis_vecs = np.concatenate((basis_Ex, basis_Ey), axis=0)

            # if enable_preconditioner:
            #    basis_vecs = (1 / precon) * basis_vecs

            # if enable_incidence_matrices:
            #    basis_vecs = dnz * basis_vecs

        # Call the eigensolver. The eigenvalues are -(neff + 1j * keff)**2
        if basis_E is None:
            vals, vecs = cls.solver_eigs(
                mat,
                num_modes,
                vec_init,
                guess_value=eig_guess,
                mode_solver_type=mode_solver_type,
                M=generalized_M,
            )
        else:
            vals, vecs = cls.solver_eigs_relative(
                mat,
                num_modes,
                vec_init,
                guess_value=eig_guess,
                mode_solver_type=mode_solver_type,
                M=generalized_M,
                basis_vecs=basis_vecs,
            )
        neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)

        # Sort by descending neff
        sort_inds = np.argsort(neff)[::-1]
        neff = neff[sort_inds]
        keff = keff[sort_inds]

        E, H = None, None
        if basis_E is None:
            if precon_right is not None:
                vecs = precon_right * vecs

            if enable_incidence_matrices:
                vecs = dnz.T * vecs

        vecs = vecs[:, sort_inds]

        # Field components from eigenvectors
        Ex = vecs[:N, :]
        Ey = vecs[N:, :]

        # Get the other field components
        h_field = qmat.dot(vecs)
        Hx = h_field[:N, :] / (1j * neff - keff)
        Hy = h_field[N:, :] / (1j * neff - keff)
        Hz = inv_mu_zz.dot(dxf.dot(Ey) - dyf.dot(Ex))
        # Ez = inv_eps_zz.dot(dxb.dot(Hy) - dyb.dot(Hx))

        # Ez = -inv_eps_zz * div^H J H_xy, while Hxy = vals^-1 * qmat * Exy;
        # Note that div^H J q_partial = 0, so Ez = -vals^-1 inv_eps_zz * div^H J q_ep Exy
        h_partial_field = q_ep.dot(vecs) / (1j * neff - keff)
        Ez = inv_eps_zz.dot(dxb.dot(h_partial_field[N:, :]) - dyb.dot(h_partial_field[:N, :]))

        # Bundle up
        E = np.stack((Ex, Ey, Ez), axis=0)
        H = np.stack((Hx, Hy, Hz), axis=0)

        # Return to standard H field units (see CEM notes for H normalization used in solver)
        H *= -1j / ETA_0

        return E, H, neff, keff

    @classmethod
    def solver_tensorial(
        cls, eps, mu, der_mats, num_modes, neff_guess, vec_init, mat_precision, direction
    ):
        """EM eigenmode solver assuming ``eps`` or ``mu`` have off-diagonal elements."""

        mode_solver_type = "tensorial"
        N = eps.shape[-1]
        dxf, dxb, dyf, dyb = der_mats

        # Compute all blocks of the matrix for diagonalization
        inv_eps_zz = sp.spdiags(1 / eps[2, 2, :], [0], N, N)
        inv_mu_zz = sp.spdiags(1 / mu[2, 2, :], [0], N, N)
        axax = -dxf.dot(sp.spdiags(eps[2, 0, :] / eps[2, 2, :], [0], N, N)) - sp.spdiags(
            mu[1, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dyf)
        axay = -dxf.dot(sp.spdiags(eps[2, 1, :] / eps[2, 2, :], [0], N, N)) + sp.spdiags(
            mu[1, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dxf)
        axbx = -dxf.dot(inv_eps_zz).dot(dyb) + sp.spdiags(
            mu[1, 0, :] - mu[1, 2, :] * mu[2, 0, :] / mu[2, 2, :], [0], N, N
        )
        axby = dxf.dot(inv_eps_zz).dot(dxb) + sp.spdiags(
            mu[1, 1, :] - mu[1, 2, :] * mu[2, 1, :] / mu[2, 2, :], [0], N, N
        )
        ayax = -dyf.dot(sp.spdiags(eps[2, 0, :] / eps[2, 2, :], [0], N, N)) + sp.spdiags(
            mu[0, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dyf)
        ayay = -dyf.dot(sp.spdiags(eps[2, 1, :] / eps[2, 2, :], [0], N, N)) - sp.spdiags(
            mu[0, 2, :] / mu[2, 2, :], [0], N, N
        ).dot(dxf)
        aybx = -dyf.dot(inv_eps_zz).dot(dyb) + sp.spdiags(
            -mu[0, 0, :] + mu[0, 2, :] * mu[2, 0, :] / mu[2, 2, :], [0], N, N
        )
        ayby = dyf.dot(inv_eps_zz).dot(dxb) + sp.spdiags(
            -mu[0, 1, :] + mu[0, 2, :] * mu[2, 1, :] / mu[2, 2, :], [0], N, N
        )
        bxbx = -dxb.dot(sp.spdiags(mu[2, 0, :] / mu[2, 2, :], [0], N, N)) - sp.spdiags(
            eps[1, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dyb)
        bxby = -dxb.dot(sp.spdiags(mu[2, 1, :] / mu[2, 2, :], [0], N, N)) + sp.spdiags(
            eps[1, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dxb)
        bxax = -dxb.dot(inv_mu_zz).dot(dyf) + sp.spdiags(
            eps[1, 0, :] - eps[1, 2, :] * eps[2, 0, :] / eps[2, 2, :], [0], N, N
        )
        bxay = dxb.dot(inv_mu_zz).dot(dxf) + sp.spdiags(
            eps[1, 1, :] - eps[1, 2, :] * eps[2, 1, :] / eps[2, 2, :], [0], N, N
        )
        bybx = -dyb.dot(sp.spdiags(mu[2, 0, :] / mu[2, 2, :], [0], N, N)) + sp.spdiags(
            eps[0, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dyb)
        byby = -dyb.dot(sp.spdiags(mu[2, 1, :] / mu[2, 2, :], [0], N, N)) - sp.spdiags(
            eps[0, 2, :] / eps[2, 2, :], [0], N, N
        ).dot(dxb)
        byax = -dyb.dot(inv_mu_zz).dot(dyf) + sp.spdiags(
            -eps[0, 0, :] + eps[0, 2, :] * eps[2, 0, :] / eps[2, 2, :], [0], N, N
        )
        byay = dyb.dot(inv_mu_zz).dot(dxf) + sp.spdiags(
            -eps[0, 1, :] + eps[0, 2, :] * eps[2, 1, :] / eps[2, 2, :], [0], N, N
        )

        mat = sp.bmat(
            [
                [axax, axay, axbx, axby],
                [ayax, ayay, aybx, ayby],
                [bxax, bxay, bxbx, bxby],
                [byax, byay, bybx, byby],
            ]
        )

        # The eigenvalues for the matrix above are 1j * (neff + 1j * keff)
        # Multiply the matrix by -1j, so that eigenvalues are (neff + 1j * keff)
        mat *= -1j

        # change matrix sign for backward direction
        if direction == "-":
            mat *= -1

        # Cast matrix to target data type
        mat_dtype = cls.matrix_data_type(eps, mu, der_mats, mat_precision, is_tensorial=True)
        mat = cls.type_conversion(mat, mat_dtype)

        # Trim small values in single precision case
        if mat_precision == "single":
            cls.trim_small_values(mat, tol=fp_eps)

        # Casting starting vector to target data type
        vec_init = cls.type_conversion(vec_init, mat_dtype)

        # Starting eigenvalue guess in target data type
        eig_guess = cls.type_conversion(np.array([neff_guess]), mat_dtype)[0]

        # Call the eigensolver.
        vals, vecs = cls.solver_eigs(
            mat,
            num_modes,
            vec_init,
            guess_value=eig_guess,
            mode_solver_type=mode_solver_type,
        )
        neff, keff = cls.eigs_to_effective_index(vals, mode_solver_type)
        # Sort by descending real part
        sort_inds = np.argsort(neff)[::-1]
        neff = neff[sort_inds]
        keff = keff[sort_inds]
        vecs = vecs[:, sort_inds]

        # Field components from eigenvectors
        Ex = vecs[:N, :]
        Ey = vecs[N : 2 * N, :]
        Hx = vecs[2 * N : 3 * N, :]
        Hy = vecs[3 * N :, :]

        # Get the other field components
        hxy_term = (-mu[2, 0, :] * Hx.T - mu[2, 1, :] * Hy.T).T
        Hz = inv_mu_zz.dot(dxf.dot(Ey) - dyf.dot(Ex) + hxy_term)
        exy_term = (-eps[2, 0, :] * Ex.T - eps[2, 1, :] * Ey.T).T
        Ez = inv_eps_zz.dot(dxb.dot(Hy) - dyb.dot(Hx) + exy_term)

        # Bundle up
        E = np.stack((Ex, Ey, Ez), axis=0)
        H = np.stack((Hx, Hy, Hz), axis=0)

        # Return to standard H field units (see CEM notes for H normalization used in solver)
        # The minus sign here is suspicious, need to check how modes are used in Mode objects
        H *= -1j / ETA_0

        return E, H, neff, keff

    @classmethod
    def solver_eigs(
        cls,
        mat,
        num_modes,
        vec_init,
        guess_value=1.0,
        M=None,
        **kwargs,
    ):
        """Find ``num_modes`` eigenmodes of ``mat`` cloest to ``guess_value``.

        Parameters
        ----------
        mat : scipy.sparse matrix
            Square matrix for diagonalization.
        num_modes : int
            Number of eigenmodes to compute.
        guess_value : float, optional
        """

        values, vectors = spl.eigs(
            mat, k=num_modes, sigma=guess_value, tol=TOL_EIGS, v0=vec_init, M=M
        )

        # for i, eig_i in enumerate(values):
        #     vec = vectors[:, i]
        #     rhs = vec
        #     if M is not None:
        #         rhs = M @ rhs
        #     eig_from_vec = (vec.T @ (mat @ vec)) / (vec.T @ rhs)
        #     residue = np.linalg.norm(mat @ vec - eig_i * rhs) / np.linalg.norm(vec)
        #     print(
        #         f"{i}-th eigenvalue: {eig_i}, referred from eigenvectors: {eig_from_vec}, relative residual: {residue}."
        #     )
        return values, vectors

    @classmethod
    def solver_eigs_relative(
        cls,
        mat,
        num_modes,
        vec_init,
        guess_value=1.0,
        M=None,
        basis_vecs=None,
        **kwargs,
    ):
        """Find ``num_modes`` eigenmodes of ``mat`` cloest to ``guess_value``.

        Parameters
        ----------
        mat : scipy.sparse matrix
            Square matrix for diagonalization.
        num_modes : int
            Number of eigenmodes to compute.
        guess_value : float, optional
        """

        basis, _ = np.linalg.qr(basis_vecs)
        mat_basis = np.conj(basis.T) @ mat @ basis
        values, coeffs = linalg.eig(mat_basis)
        vectors = None
        vectors = basis @ coeffs
        return values, vectors

    @classmethod
    def isinstance_complex(cls, vec_or_mat, tol=TOL_COMPLEX):
        """Check if a numpy array or scipy csr_matrix has complex component by looking at
        norm(x.imag)/norm(x)>TOL_COMPLEX

        Parameters
        ----------
        vec_or_mat : Union[np.ndarray, sp.csr_matrix]
        """

        if isinstance(vec_or_mat, np.ndarray):
            return np.linalg.norm(vec_or_mat.imag) / (np.linalg.norm(vec_or_mat) + fp_eps) > tol
        if isinstance(vec_or_mat, sp.csr_matrix):
            mat_norm = spl.norm(vec_or_mat)
            mat_imag_norm = spl.norm(vec_or_mat.imag)
            return mat_imag_norm / (mat_norm + fp_eps) > tol

        raise RuntimeError("Variable type should be either numpy array or scipy csr_matrix.")

    @classmethod
    def type_conversion(cls, vec_or_mat, new_dtype):
        """Convert vec_or_mat to new_type.

        Parameters
        ----------
        vec_or_mat : Union[np.ndarray, sp.csr_matrix]
            vec or mat to be converted.
        new_dtype : Union[np.complex128, np.complex64, np.float64, np.float32]
            Final type of vec or mat

        Returns
        -------
        converted_vec_or_mat : Union[np.ndarray, sp.csr_matrix]
        """

        if new_dtype in {np.complex128, np.complex64}:
            return vec_or_mat.astype(new_dtype)
        if new_dtype in {np.float64, np.float32}:
            converted_vec_or_mat = vec_or_mat.real
            return converted_vec_or_mat.astype(new_dtype)

        raise RuntimeError("Unsupported new_type.")

    @classmethod
    def set_initial_vec(cls, Nx, Ny, is_tensorial=False):
        """Set initial vector for eigs:
        1) The field at x=0 and y=0 boundaries are set to 0. This should be
        the case for PEC boundaries, but wouldn't hurt for non-PEC boundary;
        2) The vector is np.complex128 by default, and will be converted to
        appropriate type afterwards.

        Parameters
        ----------
        Nx : int
            Number of grids along x-direction.
        Ny : int
            Number of grids along y-direction.
        is_tensorial : bool
            diagonal or tensorial eigenvalue problem.
        """

        # The size of the vector is len_multiplier * Nx * Ny
        len_multiplier = 2
        if is_tensorial:
            len_multiplier *= 2

        # Initialize the vector
        size = (Nx, Ny, len_multiplier)
        rng = np.random.default_rng(0)
        vec_init = rng.random(size) + 1j * rng.random(size)

        # Set values at the boundary to be 0
        if Nx > 1:
            vec_init[0, :, :] = 0
        if Ny > 1:
            vec_init[:, 0, :] = 0

        # Concatenate the vector appropriately
        vec_init = np.vstack(vec_init)
        return vec_init.flatten("F")

    @classmethod
    def eigs_to_effective_index(cls, eig_list: Numpy, mode_solver_type: ModeSolverType):
        """Convert obtained eigenvalues to n_eff and k_eff.

        Parameters
        ----------
        eig_list : Numpy
            Array of eigenvalues
        mode_solver_type : ModeSolverType
            The type of mode solver problems

        Returns
        -------
        Tuple[Numpy, Numpy]
            n_eff and k_eff
        """
        if eig_list.size == 0:
            raise RuntimeError("Could not find any eigenmodes for this waveguide.")

        # for tensorial type, it's simply (neff + 1j * keff)
        if mode_solver_type == "tensorial":
            return np.real(eig_list), np.imag(eig_list)

        # for diagonal type, eigenvalues are -(neff + 1j * keff)**2
        if mode_solver_type == "diagonal":
            sqrt_eig_list = np.emath.sqrt(-eig_list + 0j)
            return np.real(sqrt_eig_list), np.imag(sqrt_eig_list)

        raise RuntimeError(f"Unidentified 'mode_solver_type={mode_solver_type}'.")

    @staticmethod
    def format_medium_data(mat_data):
        """
        mat_data can be either permittivity or permeability. It's either a single 2D array
        defining the relative property in the cross-section, or nine 2D arrays defining
        the property at the E(H)x, E(H)y, and E(H)z locations of the Yee grid in the order
        xx, xy, xz, yx, yy, yz, zx, zy, zz.
        """
        if isinstance(mat_data, Numpy):
            return (mat_data[i, :, :] for i in range(9))
        if len(mat_data) == 9:
            return (np.copy(e) for e in mat_data)
        raise ValueError("Wrong input to mode solver pemittivity/permeability!")

    @staticmethod
    def split_curl_field_postprocess(split_curl, E):
        """E has the shape (3, N, num_modes)"""
        _, Nx, Ny = split_curl.shape
        field_shape = E.shape

        # set a dummy value of split curl inside PEC to avoid division by 0 warning (it's 0/0, since
        # E field inside PEC is also 0); then by the end, zero out E inside PEC again just to be safe.
        outside_pec = ~np.isclose(split_curl, 0)
        split_curl_scaling = np.copy(split_curl)
        split_curl_scaling[~outside_pec] = 1.0

        E = E.reshape(3, Nx, Ny, field_shape[-1])
        E /= split_curl_scaling[:, :, :, np.newaxis]
        E *= outside_pec[:, :, :, np.newaxis]
        E = E.reshape(field_shape)
        return E

    @staticmethod
    def make_pml_invariant(Nxy, tensor, num_pml):
        """For a given epsilon or mu tensor of shape ``(3, 3, Nx, Ny)``, and ``num_pml`` pml layers
        along ``x`` and ``y``, make all the tensor values in the PML equal by replicating the first
        pixel into the PML."""

        Nx, Ny = Nxy
        new_ten = tensor.reshape((3, 3, Nx, Ny))
        new_ten[:, :, : num_pml[0], :] = new_ten[:, :, num_pml[0], :][:, :, None, :]
        new_ten[:, :, Nx - num_pml[0] + 1 :, :] = new_ten[:, :, Nx - num_pml[0], :][:, :, None, :]
        new_ten[:, :, :, : num_pml[1]] = new_ten[:, :, :, num_pml[1]][:, :, :, None]
        new_ten[:, :, :, Ny - num_pml[1] + 1 :] = new_ten[:, :, :, Ny - num_pml[1]][:, :, :, None]
        return new_ten.reshape((3, 3, -1))

    @staticmethod
    def split_curl_field_postprocess_inverse(split_curl, E):
        """E has the shape (3, N, num_modes)"""
        raise RuntimeError("Split curl not yet implemented for relative mode solver.")

    @staticmethod
    def mode_plane_contain_good_conductor(material_response) -> bool:
        """Find out if epsilon on the modal plane contain good conductors whose permittivity
        or permeability value is very large.
        """
        if material_response is None:
            return False
        return np.any(np.abs(material_response) > GOOD_CONDUCTOR_THRESHOLD * np.abs(pec_val))


def compute_modes(*args, **kwargs) -> Tuple[Numpy, Numpy, str]:
    """A wrapper around ``EigSolver.compute_modes``, which is used in ``ModeSolver``."""
    return EigSolver.compute_modes(*args, **kwargs)



"""Objects that define how data is recorded from simulation."""

from tidy3d.components.tcad.monitors.abstract import HeatChargeMonitor


class TemperatureMonitor(HeatChargeMonitor):
    """Temperature monitor."""



import inspect

import torch
from autograd import make_vjp
from autograd.extend import vspace


def to_torch(fun):
    """
    Converts an autograd function to a PyTorch function.

    Parameters
    ----------
    fun : callable
        The autograd function to be converted.

    Returns
    -------
    callable
        A PyTorch function that can be used with PyTorch tensors and supports
        autograd differentiation.

    Examples
    --------
    >>> import autograd.numpy as anp
    >>> import torch
    >>> from tidy3d.plugins.pytorch.wrapper import to_torch
    >>>
    >>> @to_torch
    ... def f(x):
    ...     return anp.sum(anp.sin(x))
    >>>
    >>> x = torch.tensor([0.0, anp.pi / 2, anp.pi], requires_grad=True)
    >>> val = f(x)
    >>> val.backward()
    >>> torch.allclose(x.grad, torch.cos(x))
    True
    """
    sig = inspect.signature(fun)

    class _Wrapper(torch.autograd.Function):
        """A `torch.autograd.Function` wrapper for the autograd function `fun`.

        See Also
        --------
        `PyTorch Autograd Function Documentation <https://pytorch.org/docs/stable/autograd.html#function>`_
        """

        @staticmethod
        def forward(ctx, *args):
            numpy_args = []
            grad_argnums = []

            # assume that all tensors are on the same device (cpu by default)
            device = torch.device("cpu")

            for idx, arg in enumerate(args):
                if torch.is_tensor(arg):
                    numpy_args.append(arg.detach().cpu().numpy())
                    device = arg.device
                    if arg.requires_grad:
                        grad_argnums.append(idx)
                else:
                    numpy_args.append(arg)

            # note: can't support differentiating w.r.t. keyword-only arguments because
            # autograd's unary_to_nary decorator passes all function arguments as positional
            _vjp = make_vjp(fun, argnum=grad_argnums)
            vjp, ans = _vjp(*numpy_args)

            ctx.vjp = vjp
            ctx.device = device
            ctx.num_args = len(args)
            ctx.grad_argnums = grad_argnums

            return torch.as_tensor(ans, device=device)

        @staticmethod
        def backward(ctx, grad_output):
            _grads = ctx.vjp(vspace(grad_output.detach().cpu().numpy()).ones())
            grads = [None] * ctx.num_args
            for idx, grad in zip(ctx.grad_argnums, _grads):
                grads[idx] = torch.as_tensor(grad, device=ctx.device) * grad_output
            return tuple(grads)

    def apply(*args, **kwargs):
        # we bind the full function signature including defaults so that we can pass
        # all values as positional since torch.autograd.Function.apply only accepts
        # positional arguments
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return _Wrapper.apply(*bound_args.arguments.values())

    return apply



# Adds some functionality to the autograd arraybox and related autograd patches
# NOTE: we do not subclass ArrayBox since that would break autograd's internal checks

import importlib
from typing import Any, Callable, Dict, List, Tuple

import autograd.numpy as anp
from autograd.extend import VJPNode, defjvp, register_notrace
from autograd.numpy.numpy_boxes import ArrayBox
from autograd.numpy.numpy_wrapper import _astype

TidyArrayBox = ArrayBox  # NOT a subclass

_autograd_module_cache = {}  # cache for imported autograd modules

register_notrace(VJPNode, anp.full_like)

defjvp(
    _astype,
    lambda g, ans, A, dtype, order="K", casting="unsafe", subok=True, copy=True: _astype(g, dtype),
)

anp.astype = _astype
anp.permute_dims = anp.transpose


@classmethod
def from_arraybox(cls, box: ArrayBox) -> TidyArrayBox:
    """Construct a TidyArrayBox from an ArrayBox."""
    return cls(box._value, box._trace, box._node)


def __array_function__(
    self: Any,
    func: Callable,
    types: List[Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
) -> Any:
    """
    Handle the dispatch of NumPy functions to autograd's numpy implementation.

    Parameters
    ----------
    self : Any
        The instance of the class.
    func : Callable
        The NumPy function being called.
    types : List[Any]
        The types of the arguments that implement __array_function__.
    args : Tuple[Any, ...]
        The positional arguments to the function.
    kwargs : Dict[str, Any]
        The keyword arguments to the function.

    Returns
    -------
    Any
        The result of the function call, or NotImplemented.

    Raises
    ------
    NotImplementedError
        If the function is not implemented in autograd.numpy.

    See Also
    --------
    https://numpy.org/doc/stable/reference/arrays.classes.html#numpy.class.__array_function__
    """
    if not all(t in TidyArrayBox.type_mappings for t in types):
        return NotImplemented

    module_name = func.__module__

    if module_name.startswith("numpy"):
        anp_module_name = "autograd." + module_name
    else:
        return NotImplemented

    # Use the cached module if available
    anp_module = _autograd_module_cache.get(anp_module_name)
    if anp_module is None:
        try:
            anp_module = importlib.import_module(anp_module_name)
            _autograd_module_cache[anp_module_name] = anp_module
        except ImportError:
            return NotImplemented

    f = getattr(anp_module, func.__name__, None)
    if f is None:
        return NotImplemented

    if f.__name__ == "nanmean":  # somehow xarray always dispatches to nanmean
        f = anp.mean
        kwargs.pop("dtype", None)  # autograd mean vjp doesn't support dtype

    return f(*args, **kwargs)


def __array_ufunc__(
    self: Any,
    ufunc: Callable,
    method: str,
    *inputs: Any,
    **kwargs: Dict[str, Any],
) -> Any:
    """
    Handle the dispatch of NumPy ufuncs to autograd's numpy implementation.

    Parameters
    ----------
    self : Any
        The instance of the class.
    ufunc : Callable
        The universal function being called.
    method : str
        The method of the ufunc being called.
    inputs : Any
        The input arguments to the ufunc.
    kwargs : Dict[str, Any]
        The keyword arguments to the ufunc.

    Returns
    -------
    Any
        The result of the ufunc call, or NotImplemented.

    See Also
    --------
    https://numpy.org/doc/stable/reference/arrays.classes.html#numpy.class.__array_ufunc__
    """
    if method != "__call__":
        return NotImplemented

    ufunc_name = ufunc.__name__

    anp_ufunc = getattr(anp, ufunc_name, None)
    if anp_ufunc is not None:
        return anp_ufunc(*inputs, **kwargs)

    return NotImplemented


def item(self):
    if self.size != 1:
        raise ValueError("Can only convert an array of size 1 to a scalar")
    return anp.ravel(self)[0]


TidyArrayBox._tidy = True
TidyArrayBox.from_arraybox = from_arraybox
TidyArrayBox.__array_namespace__ = lambda self, *, api_version=None: anp
TidyArrayBox.__array_ufunc__ = __array_ufunc__
TidyArrayBox.__array_function__ = __array_function__
TidyArrayBox.__repr__ = str
TidyArrayBox.real = property(anp.real)
TidyArrayBox.imag = property(anp.imag)
TidyArrayBox.conj = anp.conj
TidyArrayBox.item = item



"""Deprecated module"""

from ...log import log

log.warning(
    "The module 'plugins.dispersion.fit_web' has been deprecated in favor of "
    "'plugins.dispersion.web' and will be removed in future versions."
)



"""Tool for generating an S matrix automatically from a Tidy3d simulation and modal port definitions."""

# TODO: The names "ComponentModeler" and "Port" should be changed to "ModalComponentModeler" and
# "ModalPort" to explicitly differentiate these from "TerminalComponentModeler" and "LumpedPort".
from __future__ import annotations

from typing import Dict, List, Optional, Tuple

import numpy as np
import pydantic.v1 as pd

from ....components.base import cached_property
from ....components.data.sim_data import SimulationData
from ....components.monitor import ModeMonitor
from ....components.simulation import Simulation
from ....components.source.field import ModeSource
from ....components.source.time import GaussianPulse
from ....components.types import Ax, Complex
from ....components.viz import add_ax_if_none, equal_aspect
from ....exceptions import SetupError
from ....web.api.container import BatchData
from ..ports.modal import ModalPortDataArray, Port
from .base import FWIDTH_FRAC, AbstractComponentModeler

MatrixIndex = Tuple[str, pd.NonNegativeInt]  # the 'i' in S_ij
Element = Tuple[MatrixIndex, MatrixIndex]  # the 'ij' in S_ij


class ComponentModeler(AbstractComponentModeler):
    """
    Tool for modeling devices and computing scattering matrix elements.

    .. TODO missing basic example

    See Also
    --------

    **Notebooks**
        * `Computing the scattering matrix of a device <../../notebooks/SMatrix.html>`_
    """

    ports: Tuple[Port, ...] = pd.Field(
        (),
        title="Ports",
        description="Collection of ports describing the scattering matrix elements. "
        "For each input mode, one simulation will be run with a modal source.",
    )

    element_mappings: Tuple[Tuple[Element, Element, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Mapping between elements of the scattering matrix, "
        "as specified by pairs of ``(port name, mode index)`` matrix indices, where the "
        "first element of the pair is the output and the second element of the pair is the input."
        "Each item of ``element_mappings`` is a tuple of ``(element1, element2, c)``, where "
        "the scattering matrix ``Smatrix[element2]`` is set equal to ``c * Smatrix[element1]``."
        "If all elements of a given column of the scattering matrix are defined by "
        " ``element_mappings``, the simulation corresponding to this column "
        "is skipped automatically.",
    )

    run_only: Optional[Tuple[MatrixIndex, ...]] = pd.Field(
        None,
        title="Run Only",
        description="If given, a tuple of matrix indices, specified by (:class:`.Port`, ``int``),"
        " to run only, excluding the other rows from the scattering matrix. "
        "If this option is used, "
        "the data corresponding to other inputs will be missing in the resulting matrix.",
    )
    """Finally, to exclude some rows of the scattering matrix, one can supply a ``run_only`` parameter to the
    :class:`ComponentModeler`. ``run_only`` contains the scattering matrix indices that the user wants to run as a
    source. If any indices are excluded, they will not be run."""

    verbose: bool = pd.Field(
        False,
        title="Verbosity",
        description="Whether the :class:`.ComponentModeler` should print status and progressbars.",
    )

    callback_url: str = pd.Field(
        None,
        title="Callback URL",
        description="Http PUT url to receive simulation finish event. "
        "The body content is a json file with fields "
        "``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.",
    )

    @pd.validator("simulation", always=True)
    def _sim_has_no_sources(cls, val):
        """Make sure simulation has no sources as they interfere with tool."""
        if len(val.sources) > 0:
            raise SetupError("'ComponentModeler.simulation' must not have any sources.")
        return val

    @cached_property
    def sim_dict(self) -> Dict[str, Simulation]:
        """Generate all the :class:`.Simulation` objects for the S matrix calculation."""

        sim_dict = {}
        mode_monitors = [self.to_monitor(port=port) for port in self.ports]

        for port_name, mode_index in self.matrix_indices_run_sim:
            port = self.get_port_by_name(port_name=port_name)

            port_source = self.shift_port(port=port)
            mode_source = self.to_source(port=port_source, mode_index=mode_index)

            new_mnts = list(self.simulation.monitors) + mode_monitors
            sim_copy = self.simulation.copy(update=dict(sources=[mode_source], monitors=new_mnts))
            task_name = self._task_name(port=port, mode_index=mode_index)
            sim_dict[task_name] = sim_copy
        return sim_dict

    @cached_property
    def matrix_indices_monitor(self) -> Tuple[MatrixIndex, ...]:
        """Tuple of all the possible matrix indices (port, mode_index) in the Component Modeler."""
        matrix_indices = []
        for port in self.ports:
            for mode_index in range(port.mode_spec.num_modes):
                matrix_indices.append((port.name, mode_index))
        return tuple(matrix_indices)

    @cached_property
    def matrix_indices_source(self) -> Tuple[MatrixIndex, ...]:
        """Tuple of all the source matrix indices (port, mode_index) in the Component Modeler."""
        if self.run_only is not None:
            return self.run_only
        return self.matrix_indices_monitor

    @cached_property
    def matrix_indices_run_sim(self) -> Tuple[MatrixIndex, ...]:
        """Tuple of all the source matrix indices (port, mode_index) in the Component Modeler."""

        if self.element_mappings is None or self.element_mappings == {}:
            return self.matrix_indices_source

        # all the (i, j) pairs in `S_ij` that are tagged as covered by `element_mappings`
        elements_determined_by_map = [element_out for (_, element_out, _) in self.element_mappings]

        # loop through rows of the full s matrix and record rows that still need running.
        source_indices_needed = []
        for col_index in self.matrix_indices_source:
            # loop through columns and keep track of whether each element is covered by mapping.
            matrix_elements_covered = []
            for row_index in self.matrix_indices_monitor:
                element = (row_index, col_index)
                element_covered_by_map = element in elements_determined_by_map
                matrix_elements_covered.append(element_covered_by_map)

            # if any matrix elements in row still not covered by map, a source is needed for row.
            if not all(matrix_elements_covered):
                source_indices_needed.append(col_index)

        return source_indices_needed

    @cached_property
    def port_names(self) -> Tuple[List[str], List[str]]:
        """List of port names for inputs and outputs, respectively."""

        def get_port_names(matrix_elements: Tuple[str, int]) -> List[str]:
            """Get the port names from a list of (port name, mode index)."""
            port_names = []
            for port_name, _ in matrix_elements:
                if port_name not in port_names:
                    port_names.append(port_name)
            return port_names

        port_names_in = get_port_names(self.matrix_indices_source)
        port_names_out = get_port_names(self.matrix_indices_monitor)

        return port_names_out, port_names_in

    def to_monitor(self, port: Port) -> ModeMonitor:
        """Creates a mode monitor from a given port."""
        return ModeMonitor(
            center=port.center,
            size=port.size,
            freqs=self.freqs,
            mode_spec=port.mode_spec,
            name=port.name,
        )

    def to_source(
        self, port: Port, mode_index: int, num_freqs: int = 1, **kwargs
    ) -> List[ModeSource]:
        """Creates a list of mode sources from a given port."""
        freq0 = np.mean(self.freqs)
        fdiff = max(self.freqs) - min(self.freqs)
        fwidth = max(fdiff, freq0 * FWIDTH_FRAC)
        return ModeSource(
            center=port.center,
            size=port.size,
            source_time=GaussianPulse(freq0=freq0, fwidth=fwidth),
            mode_spec=port.mode_spec,
            mode_index=mode_index,
            direction=port.direction,
            name=port.name,
            num_freqs=num_freqs,
            **kwargs,
        )

    def shift_port(self, port: Port) -> Port:
        """Generate a new port shifted by the shift amount in normal direction."""

        shift_value = self._shift_value_signed(port=port)
        center_shifted = list(port.center)
        center_shifted[port.size.index(0.0)] += shift_value
        port_shifted = port.copy(update=dict(center=center_shifted))
        return port_shifted

    @equal_aspect
    @add_ax_if_none
    def plot_sim(self, x: float = None, y: float = None, z: float = None, ax: Ax = None) -> Ax:
        """Plot a :class:`.Simulation` with all sources added for each port, for troubleshooting."""

        plot_sources = []
        for port_source in self.ports:
            mode_source_0 = self.to_source(port=port_source, mode_index=0)
            plot_sources.append(mode_source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot(x=x, y=y, z=z, ax=ax)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self, x: float = None, y: float = None, z: float = None, ax: Ax = None, **kwargs
    ) -> Ax:
        """Plot permittivity of the :class:`.Simulation` with all sources added for each port."""

        plot_sources = []
        for port_source in self.ports:
            mode_source_0 = self.to_source(port=port_source, mode_index=0)
            plot_sources.append(mode_source_0)
        sim_plot = self.simulation.copy(update=dict(sources=plot_sources))
        return sim_plot.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    def _normalization_factor(self, port_source: Port, sim_data: SimulationData) -> complex:
        """Compute the normalization amplitude based on the measured input mode amplitude."""

        port_monitor_data = sim_data[port_source.name]
        mode_index = sim_data.simulation.sources[0].mode_index

        normalize_amps = port_monitor_data.amps.sel(
            f=np.array(self.freqs),
            direction=port_source.direction,
            mode_index=mode_index,
        )

        return normalize_amps.values

    @cached_property
    def max_mode_index(self) -> Tuple[int, int]:
        """maximum mode indices for the smatrix dataset for the in and out ports, respectively."""

        def get_max_mode_indices(matrix_elements: Tuple[str, int]) -> int:
            """Get the maximum mode index for a list of (port name, mode index)."""
            return max(mode_index for _, mode_index in matrix_elements)

        max_mode_index_out = get_max_mode_indices(self.matrix_indices_monitor)
        max_mode_index_in = get_max_mode_indices(self.matrix_indices_source)

        return max_mode_index_out, max_mode_index_in

    def _construct_smatrix(self) -> ModalPortDataArray:
        """Post process :class:`.BatchData` to generate scattering matrix."""
        return self._internal_construct_smatrix(batch_data=self.batch_data)

    def _internal_construct_smatrix(self, batch_data: BatchData) -> ModalPortDataArray:
        """Post process :class:`.BatchData` to generate scattering matrix, for internal use only."""

        max_mode_index_out, max_mode_index_in = self.max_mode_index
        num_modes_out = max_mode_index_out + 1
        num_modes_in = max_mode_index_in + 1
        port_names_out, port_names_in = self.port_names

        values = np.zeros(
            (len(port_names_out), len(port_names_in), num_modes_out, num_modes_in, len(self.freqs)),
            dtype=complex,
        )
        coords = dict(
            port_out=port_names_out,
            port_in=port_names_in,
            mode_index_out=range(num_modes_out),
            mode_index_in=range(num_modes_in),
            f=np.array(self.freqs),
        )
        s_matrix = ModalPortDataArray(values, coords=coords)

        # loop through source ports
        for col_index in self.matrix_indices_run_sim:
            port_name_in, mode_index_in = col_index
            port_in = self.get_port_by_name(port_name=port_name_in)

            sim_data = batch_data[self._task_name(port=port_in, mode_index=mode_index_in)]

            for row_index in self.matrix_indices_monitor:
                port_name_out, mode_index_out = row_index
                port_out = self.get_port_by_name(port_name=port_name_out)

                # directly compute the element
                mode_amps_data = sim_data[port_out.name].copy().amps
                dir_out = "-" if port_out.direction == "+" else "+"
                amp = mode_amps_data.sel(
                    f=coords["f"], direction=dir_out, mode_index=mode_index_out
                )
                source_norm = self._normalization_factor(port_in, sim_data)
                s_matrix_elements = np.array(amp.data) / np.array(source_norm)
                s_matrix.loc[
                    dict(
                        port_in=port_name_in,
                        mode_index_in=mode_index_in,
                        port_out=port_name_out,
                        mode_index_out=mode_index_out,
                    )
                ] = s_matrix_elements

        # element can be determined by user-defined mapping
        for (row_in, col_in), (row_out, col_out), mult_by in self.element_mappings:
            port_out_from, mode_index_out_from = row_in
            port_in_from, mode_index_in_from = col_in
            coords_from = dict(
                port_in=port_in_from,
                mode_index_in=mode_index_in_from,
                port_out=port_out_from,
                mode_index_out=mode_index_out_from,
            )

            port_out_to, mode_index_out_to = row_out
            port_in_to, mode_index_in_to = col_out
            coords_to = dict(
                port_in=port_in_to,
                mode_index_in=mode_index_in_to,
                port_out=port_out_to,
                mode_index_out=mode_index_out_to,
            )
            s_matrix.loc[coords_to] = mult_by * s_matrix.loc[coords_from].values

        return s_matrix
